##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistent Data Manager"""
import binascii
import hashlib
import logging
import os
import random
import re
import socket
import struct
import threading
import time
import traceback
import uuid
from collections.abc import MutableMapping
from typing import Optional

import psycopg2
import psycopg2.errorcodes
import psycopg2.extensions
import psycopg2.extras
import transaction

import pjpersist.sqlbuilder as sb
import zope.interface
from pjpersist import interfaces, serialize
from pjpersist.querystats import QueryReport

# Flag enabling full two-phase-commit support. Note, that this requires
# postgres database to set max_prepared_transactions setting to positive value.
PJ_TWO_PHASE_COMMIT_ENABLED = False

PJ_ACCESS_LOGGING = False
# set to True to automatically create tables if they don't exist
# it is relatively expensive, so create your tables with a schema.sql
# and turn this off for production

# Enable query statistics reporting after transaction ends
PJ_ENABLE_QUERY_STATS = False

# you can register listeners to GLOBAL_QUERY_STATS_LISTENERS with
# `register_query_stats_listener`.
GLOBAL_QUERY_STATS_LISTENERS = set()

# Maximum query length to output with query log
MAX_QUERY_ARGUMENT_LENGTH = 500


PJ_AUTO_CREATE_TABLES = True

# set to True to automatically create IColumnSerialization columns
# will also create tables regardless of the PJ_AUTO_CREATE_TABLES setting
# so this is super expensive
PJ_AUTO_CREATE_COLUMNS = True


TABLE_LOG = logging.getLogger('pjpersist.table')

THREAD_NAMES = []
THREAD_COUNTERS = {}

# When a conflict error is thrown we want to ensure it gets
# handled properly at a higher level (i.e. the transaction is
# retried). In cases where is it not handled correctly we
# will have the traceback of where the conflict occurred
# for debugging later.
# Make sure you set CONFLICT_TRACEBACK_INFO.traceback to None
# after the error is processed.
CONFLICT_TRACEBACK_INFO = threading.local()
CONFLICT_TRACEBACK_INFO.traceback = None

mhash = hashlib.md5()
mhash.update(socket.gethostname().encode('utf-8'))
HOSTNAME_HASH = mhash.digest()[:3]
PID_HASH = struct.pack(">H", os.getpid() % 0xFFFF)

LOG = logging.getLogger(__name__)

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)

PREPARED_TRANSACTION_ID = object()

SERIALIZATION_ERRORS = (
    psycopg2.errorcodes.SERIALIZATION_FAILURE,
    psycopg2.errorcodes.DEADLOCK_DETECTED
)

DISCONNECTED_EXCEPTIONS = (
    psycopg2.OperationalError,
    psycopg2.InterfaceError,
)

# Hints to decide what sort of SQL command we're about to execute
# this is a VERY simple estimation whether the given SQL command
# is a read or write operation
# An exact solution would be to parse the SQL command, but that is
# A) slow,
# B) out of scope,
# C) since our code issues the commands we're in control
SQL_FIRST_WORDS = {
    'alter': 'ddl',
    'create': 'ddl',
    'delete': 'write',
    'drop': 'ddl',
    'insert': 'write',
    'update': 'write',
    'with': 'read',
    'select': 'read',
    'truncate': 'write',
    }

# Flag whether tpc_vote should call tpc_prepare when the transaction
# had no writes. See PJDataManager.tpc_vote
CALL_TPC_PREPARE_ON_NO_WRITE_TRANSACTION = True
# Flag whether PJDataManager should log on commit whether the transaction
# had writes.
LOG_READ_WRITE_TRANSACTION = False


def createId():
    # 4 bytes current time
    id = struct.pack(">i", int(time.time()))
    # 3 bytes machine
    id += HOSTNAME_HASH
    # 2 bytes pid
    id += PID_HASH
    # 1 byte thread id
    tname = threading.currentThread().name
    if tname not in THREAD_NAMES:
        THREAD_NAMES.append(tname)
    tidx = THREAD_NAMES.index(tname)
    id += struct.pack(">B", tidx & 0xFF)
    # 2 bytes counter
    THREAD_COUNTERS.setdefault(tidx, random.randint(0, 0xFFFF))
    THREAD_COUNTERS[tidx] += 1
    id += struct.pack(">H", THREAD_COUNTERS[tidx] & 0xFFFF)
    return binascii.hexlify(id).decode('ascii')


class Json(psycopg2.extras.Json):
    """In logs, we want to have the JSON value not just Json object at <>"""
    def __repr__(self):
        if PJ_ACCESS_LOGGING:
            try:
                s = self.dumps(self.adapted)
            except:
                s = 'exception'
            return '<%s %s>' % (self.__class__.__name__, s)
        else:
            return '<%s>' % (self.__class__.__name__, )


class PJPersistCursor(psycopg2.extras.DictCursor):
    def __init__(self, datamanager, flush, *args, **kwargs):
        super(PJPersistCursor, self).__init__(*args, **kwargs)
        self.datamanager = datamanager
        self.flush = flush

    def log_query(self, sql, args, duration):

        txn = transaction.get()
        txn = '%i - %s' % (id(txn), txn.description)

        TABLE_LOG.debug(
            "%s,\n args:%r,\n TXN:%s,\n time:%sms",
            sql, args, txn, duration*1000)

    def _autoCreateTables(self, sql, args, beacon):
        # XXX: need to set a savepoint, just in case the real execute
        #      fails, it would take down all further commands
        super(PJPersistCursor, self).execute("SAVEPOINT before_execute;")

        try:
            return self._execute_and_log(sql, args)
        except psycopg2.Error as e:
            # XXX: ugly: we're creating here missing tables on the fly
            msg = str(e)
            TABLE_LOG.debug("%s %r failed with %s", sql, args, msg)
            # if the exception message matches
            m = re.search('relation "(.*?)" does not exist', msg)
            if m:
                # need to rollback to the above savepoint, otherwise
                # PG would just ignore any further command
                super(PJPersistCursor, self).execute(
                    "ROLLBACK TO SAVEPOINT before_execute;")

                # we extract the tableName from the exception message
                tableName = m.group(1)

                self.datamanager._create_doc_table(
                    self.datamanager.database, tableName)

                try:
                    return self._execute_and_log(sql, args)
                except psycopg2.Error:
                    # Join the transaction, because failed queries require
                    # aborting the transaction.
                    self.datamanager._join_txn()
            # Join the transaction, because failed queries require
            # aborting the transaction.
            self.datamanager._join_txn()
            self.datamanager._doom_txn()
            check_for_conflict(e, sql, beacon=beacon)
            check_for_disconnect(e, sql, beacon=beacon)
            # otherwise let it fly away
            raise

    def execute(self, sql, args=None, beacon=None, flush_hint=None):
        """execute a SQL statement
        sql - SQL string or SQLBuilder expression
        args - optional list of args for the SQL string
        beacon - optional unique identifier for the statement to help
                 debugging errors. Going to be added to all possible
                 exceptions and log entries
        flush_hint - list of tables to flush before querying database
                     or None to flush all
        """
        # Convert SQLBuilder object to string
        if not isinstance(sql, str):
            sql = sql.__sqlrepr__('postgres')

        # Determine SQL command type (read/write), well sort of
        # See also comments on SQL_FIRST_WORDS
        firstWord = sql.strip().split()[0].lower()
        # By default we opt for 'write' to be on the safe side
        sqlCommandType = SQL_FIRST_WORDS.get(firstWord, 'write')
        if sqlCommandType in ('write', 'ddl'):
            self.datamanager.setDirty()

        if self.flush and sqlCommandType == 'read' and flush_hint != []:
            # Flush the data manager before any select.
            # We do this to have the written data available for queries
            self.datamanager.flush(flush_hint=flush_hint)

        # XXX: Optimization opportunity to store returned JSONB docs in the
        # cache of the data manager. (SR)

        if PJ_AUTO_CREATE_TABLES:
            self._autoCreateTables(sql, args, beacon)
        else:
            try:
                # otherwise just execute the given sql
                return self._execute_and_log(sql, args)
            except psycopg2.Error as e:
                # Join the transaction, because failed queries require
                # aborting the transaction.
                self.datamanager._join_txn()
                self.datamanager._doom_txn()
                check_for_conflict(e, sql, beacon=beacon)
                check_for_disconnect(e, sql, beacon=beacon)
                raise

    def _sanitize_arg(self, arg):
        r = repr(arg)
        if len(r) > MAX_QUERY_ARGUMENT_LENGTH:
            r = r[:MAX_QUERY_ARGUMENT_LENGTH] + "..."
            return r
        return arg

    def _execute_and_log(self, sql, args):
        # Very useful logging of every SQL command with traceback to code.
        __traceback_info__ = (self.datamanager.database, sql, args)
        started = time.time()
        try:
            res = super(PJPersistCursor, self).execute(sql, args)
        finally:
            duration = time.time() - started
            db = self.datamanager.database

            debug = (PJ_ACCESS_LOGGING or PJ_ENABLE_QUERY_STATS)

            if debug:
                saneargs = [self._sanitize_arg(a) for a in args] \
                    if args else args
            else:
                # We don't want to do expensive sanitization in prod mode
                saneargs = args

            if PJ_ACCESS_LOGGING:
                self.log_query(sql, saneargs, duration)

            if PJ_ENABLE_QUERY_STATS:
                self.datamanager._query_report.record(
                    sql, saneargs, duration, db)

            for rep in GLOBAL_QUERY_STATS_LISTENERS:
                rep.record(sql, saneargs, duration, db)
        return res


def check_for_conflict(e, sql, beacon=None):
    """Check whether exception indicates serialization failure and raise
    ConflictError in this case.

    Serialization failures are denoted by postgres codes:
        40001 - serialization_failure
        40P01 - deadlock_detected

    e - exception caught
    sql - SQL string that was executed when the exception occurred
    beacon - optional unique identifier for the statement to help
             debugging errors. Going to be added to all possible
             exceptions and log entries
    """
    if e.pgcode in SERIALIZATION_ERRORS:
        if beacon is None:
            beacon = uuid.uuid4()
        beacon = 'Beacon: %s' % beacon
        CONFLICT_TRACEBACK_INFO.traceback = (
            traceback.format_stack() + [beacon])
        LOG.warning("Conflict detected with code %s sql: %s, %s",
                    e.pgcode, sql, beacon)
        raise interfaces.ConflictError(str(e).strip(), beacon, sql)


def check_for_disconnect(e, sql, beacon=None):
    """Check whether exception indicates a database disconnected error
    and raise DatabaseDisconnected in this case
    Note, DB disconnect detection is NOT easy

    e - exception caught
    sql - SQL string that was executed when the exception occurred
    beacon - optional unique identifier for the statement to help
             debugging errors. Going to be added to all possible
             exceptions and log entries
    """
    if isinstance(e, DISCONNECTED_EXCEPTIONS):
        # XXX: having an exception of the above type might NOT be
        #      an exact sign of a DB disconnect, but as it looks we
        #      don't have a better chance to figure
        if beacon is None:
            beacon = uuid.uuid4()
        beacon = 'Beacon: %s' % beacon
        LOG.warning(
            "Caught exception %r, reraising as DatabaseDisconnected %s",
            e, beacon)
        raise interfaces.DatabaseDisconnected(str(e).strip(), beacon, sql)


class Root(MutableMapping):

    table = 'persistence_root'

    def __init__(self, jar, table=None):
        self._jar = jar
        if table is not None:
            self.table = table
        if PJ_AUTO_CREATE_TABLES:
            self._init_table()

    def _init_table(self):
        with self._jar.getCursor(False) as cur:
            cur.execute(
                "SELECT * FROM information_schema.tables where table_name=%s",
                (self.table,))
            if cur.rowcount:
                return

            LOG.info("Creating table %s" % self.table)
            cur.execute('''
                CREATE TABLE %s (
                    id SERIAL PRIMARY KEY,
                    name TEXT,
                    dbref TEXT[])
                ''' % self.table)
            self._jar._conn.commit()

    def __getitem__(self, key):
        with self._jar.getCursor(False) as cur:
            tbl = sb.Table(self.table)
            cur.execute(
                sb.Select(sb.Field(self.table, 'dbref'), tbl.name == key))
            if not cur.rowcount:
                raise KeyError(key)
            db, tbl, id = cur.fetchone()['dbref']
            dbref = serialize.DBRef(tbl, id, db)
            return self._jar.load(dbref)

    def __setitem__(self, key, value):
        dbref = self._jar.insert(value)
        if self.get(key) is not None:
            del self[key]
        with self._jar.getCursor(False) as cur:
            cur.execute(
                'INSERT INTO %s (name, dbref) VALUES (%%s, %%s)' % self.table,
                (key, list(dbref.as_tuple()))
                )

    def __delitem__(self, key):
        self._jar.remove(self[key])
        with self._jar.getCursor(False) as cur:
            tbl = sb.Table(self.table)
            cur.execute(sb.Delete(self.table, tbl.name == key))

    def keys(self):
        with self._jar.getCursor(False) as cur:
            cur.execute(sb.Select(sb.Field(self.table, 'name')))
            return [doc['name'] for doc in cur.fetchall()]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())


@zope.interface.implementer(interfaces.IPJDataManager)
class PJDataManager(object):

    root = None

    # Data manager is completely new. NOTE: It is important to leave this
    # property on class level and not add it to constructor, because
    # constructor is called to "reset" the data manager
    _pristine = True

    # Flag showing whether there is any write in the current transaction
    _dirty: bool

    _isolation_level: Optional[str]
    _readonly: Optional[bool]
    _deferrable: Optional[bool]

    def __init__(self, pool, root_table=None):
        self._pool = pool
        self._reader = serialize.ObjectReader(self)
        self._writer = serialize.ObjectWriter(self)
        self.transaction_manager = transaction.manager
        self._query_report = QueryReport()
        self._reset_data_manager()

        if self.root is None:
            # Getting Root can call self._join_txn when the table gets
            # auto-created, that calls self._begin, that might set
            # self._tpc_activated, so do this after setting self._tpc_activated
            self.root = Root(self, root_table)

    def _reset_data_manager(self):
        self._conn = None
        self.database = None
        # All of the following object lists are keys by object id. This is
        # needed when testing containment, since that can utilize `__cmp__()`
        # which can have undesired side effects. `id()` is guaranteed to not
        # use any method or state of the object itself.
        self._registered_objects = {}
        self._loaded_objects = {}
        self._inserted_objects = {}
        self._removed_objects = {}
        self._dirty = False

        self._readonly = None
        self._deferrable = None
        self._isolation_level = None

        # The latest states written to the database.
        self._latest_states = {}
        self._needs_to_join = True
        self._object_cache = {}
        self.annotations = {}

        self._txn_active = False
        self._tpc_activated = False

    def getCursor(self, flush=True):
        self._join_txn()
        def factory(*args, **kwargs):
            return PJPersistCursor(self, flush, *args, **kwargs)
        cur = self._conn.cursor(cursor_factory=factory)

        if not self._txn_active:
            # clear any traceback before starting next txn
            CONFLICT_TRACEBACK_INFO.traceback = None
            self._txn_active = True
        return cur

    def createId(self):
        return createId()

    def create_tables(self, tables):
        if isinstance(tables, str):
            tables = [tables]

        for tbl in tables:
            self._create_doc_table(self.database, tbl)

        with self.getCursor(False) as cur:
            cur.connection.commit()

    def _create_doc_table(self, database, table, extra_columns=''):
        if self.database != database:
            raise NotImplementedError(
                'Cannot store an object of a different database.',
                self.database, database)

        with self.getCursor(False) as cur:
            cur.execute(
                "SELECT * FROM information_schema.tables WHERE table_name=%s",
                (table,))
            if not cur.rowcount:
                LOG.info("Creating data table %s" % table)
                if extra_columns:
                    extra_columns += ', '
                cur.execute('''
                    CREATE TABLE %s (
                        id VARCHAR(24) NOT NULL PRIMARY KEY, %s
                        data JSONB)''' % (table, extra_columns))
                # this index helps a tiny bit with JSONB_CONTAINS queries
                cur.execute('''
                    CREATE INDEX %s_data_gin ON %s USING GIN (data);
                    ''' % (table, table))

    def _ensure_sql_columns(self, obj, table):
        # create the table required for the object, with the necessary
        # _pj_column_fields translated to SQL types
        if PJ_AUTO_CREATE_COLUMNS:
            if interfaces.IColumnSerialization.providedBy(obj):
                # XXX: exercise for later, not just create but check
                #      the columns
                # SELECT column_name
                #  FROM INFORMATION_SCHEMA.COLUMNS
                #  WHERE table_name = '<name of table>';
                columns = []
                for field in obj._pj_column_fields:
                    pgtype = serialize.PYTHON_TO_PG_TYPES[field._type]
                    columns.append("%s %s" % (field.__name__, pgtype))

                columns = ', '.join(columns)

                self._create_doc_table(self.database, table, columns)

    def _insert_doc(self, database, table, doc, id=None, column_data=None):
        # Create id if it is None.
        if id is None:
            id = self.createId()
        # Insert the document into the table.
        with self.getCursor() as cur:
            builtins = dict(id=id, data=Json(doc))
            if column_data is None:
                column_data = builtins
            else:
                column_data.update(builtins)

            columns = []
            values = []
            for colname, value in column_data.items():
                columns.append(colname)
                values.append(value)
            placeholders = ', '.join(['%s'] * len(columns))
            columns = ', '.join(columns)
            sql = "INSERT INTO %s (%s) VALUES (%s)" % (
                table, columns, placeholders)

            cur.execute(sql, tuple(values),
                        beacon="%s:%s:%s" % (database, table, id))
        return id

    def _update_doc(self, database, table, doc, id, column_data=None):
        # Insert the document into the table.
        with self.getCursor() as cur:
            builtins = dict(data=Json(doc))
            if column_data is None:
                column_data = builtins
            else:
                column_data.update(builtins)

            columns = []
            values = []
            for colname, value in column_data.items():
                columns.append(colname+'=%s')
                values.append(value)
            columns = ', '.join(columns)
            sql = "UPDATE %s SET %s WHERE id = %%s" % (table, columns)

            cur.execute(sql, tuple(values) + (id,),
                        beacon="%s:%s:%s" % (database, table, id))
        return id

    def _get_doc(self, database, table, id):
        tbl = sb.Table(table)
        with self.getCursor() as cur:
            cur.execute(sb.Select(sb.Field(table, '*'), tbl.id == id),
                        beacon="%s:%s:%s" % (database, table, id),
                        flush_hint=[table])
            res = cur.fetchone()
            return res['data'] if res is not None else None

    def _get_doc_by_dbref(self, dbref):
        return self._get_doc(dbref.database, dbref.table, dbref.id)

    def _get_doc_py_type(self, database, table, id):
        tbl = sb.Table(table)
        with self.getCursor() as cur:
            datafld = sb.Field(table, 'data')
            cur.execute(
                sb.Select(sb.JGET(datafld, interfaces.PY_TYPE_ATTR_NAME),
                          tbl.id == id),
                beacon="%s:%s:%s" % (database, table, id))
            res = cur.fetchone()
            return res[0] if res is not None else None

    def _get_table_from_object(self, obj):
        return self._writer.get_table_name(obj)

    def flush(self, flush_hint=None):
        # flush_hint contains tables that we want to flush, leaving all other
        # objects registered.
        # Now write every registered object, but make sure we write each
        # object just once.
        processed = set()
        flushed = set()
        docobject_flushed = set()
        # Make sure that we do not compute the list of flushable objects all
        # at once. While writing objects, new sub-objects might be registered
        # that also need saving.
        todo = set(self._registered_objects.keys())
        while todo:
            obj_id = todo.pop()
            obj = self._registered_objects[obj_id]
            __traceback_info__ = obj

            docobj = self._get_doc_object(obj)
            docobj_id = id(docobj)
            processed.add(obj_id)

            dbname, table = self._writer.get_table_name(docobj)
            if flush_hint and table not in flush_hint:
                continue

            flushed.add(obj_id)

            if docobj_id not in docobject_flushed:
                self._writer.store(docobj)
                docobject_flushed.add(docobj_id)

            self.setDirty()
            todo = set(self._registered_objects.keys()) - processed

        # Let's now reset all objects as if they were not modified:
        for obj_id in flushed:
            # Another nested flush already took care of this object.
            if obj_id not in self._registered_objects:
                continue
            obj = self._registered_objects[obj_id]
            obj._p_changed = False

        self._registered_objects = {
            obj_id: obj
            for obj_id, obj in self._registered_objects.items()
            if obj_id not in flushed
        }

    def _get_doc_object(self, obj):
        seen = []
        # Make sure we write the object representing a document in a
        # table and not a sub-object.
        while getattr(obj, '_p_pj_sub_object', False):
            if id(obj) in seen:
                raise interfaces.CircularReferenceError(obj)
            seen.append(id(obj))
            obj = obj._p_pj_doc_object
        return obj

    def _join_txn(self):
        if self._needs_to_join:
            self._acquire_conn()
            # once we have a working connection, we can join the transaction
            transaction = self.transaction_manager.get()
            transaction.join(self)
            self._needs_to_join = False
            self._begin(transaction)

    def _doom_txn(self):
        try:
            transaction = self.transaction_manager.get()
            transaction.doom()
            return True
        except ValueError:
            # We are not in active trasaction phase,
            # so it cannot be doomed.
            return False

    def dump(self, obj):
        res = self._writer.store(obj)
        self.setDirty()
        if id(obj) in self._registered_objects:
            obj._p_changed = False
            del self._registered_objects[id(obj)]
        return res

    def load(self, dbref, klass=None):
        self._join_txn()
        dm = self
        if dbref.database != self.database:
            # This is a reference of object from different database! We need to
            # locate the suitable data manager for this.
            dmp = zope.component.getUtility(interfaces.IPJDataManagerProvider)
            dm = dmp.get(dbref.database)
            dm._join_txn()
            assert dm.database == dbref.database, (dm.database, dbref.database)
            return dm.load(dbref, klass)

        return self._reader.get_ghost(dbref, klass)

    def reset(self):
        # we need to issue rollback on self._conn too, to get the latest
        # DB updates, not just reset PJDataManager state
        self.abort(None)

    def _acquire_conn(self):
        """Get connection from connection pool"""
        self._conn = self._pool.getconn()
        assert self._conn.status == psycopg2.extensions.STATUS_READY
        self.database = get_database_name_from_dsn(self._conn.dsn)

    def _release_conn(self, conn):
        """Release the connection after transaction is complete
        """
        if not conn.closed:
            # Set transaction options back to their default values so that next
            # transaction is not affected
            conn.reset()
        self._pool.putconn(conn)
        self._reset_data_manager()

    def insert(self, obj, oid=None):
        self._join_txn()
        if obj._p_oid is not None:
            raise ValueError('Object._p_oid is already set.', obj)
        res = self._writer.store(obj, id=oid)
        self.setDirty()
        obj._p_changed = False
        self._object_cache[hash(obj._p_oid)] = obj
        self._inserted_objects[id(obj)] = obj
        return res

    def remove(self, obj):
        if obj._p_oid is None:
            raise ValueError('Object._p_oid is None.', obj)
        # If the object is still in the ghost state, let's load it, so that we
        # have the state in case we abort the transaction later.
        if obj._p_changed is None:
            self.setstate(obj)
        # Now we remove the object from PostGreSQL.
        dbname, table = self._get_table_from_object(obj)
        with self.getCursor() as cur:
            cur.execute('DELETE FROM %s WHERE id = %%s' % table,
                        (obj._p_oid.id,),
                        beacon="%s:%s:%s" % (dbname, table, obj._p_oid.id))
            self.setDirty()
        if hash(obj._p_oid) in self._object_cache:
            del self._object_cache[hash(obj._p_oid)]

        # Edge case: The object was just added in this transaction.
        if id(obj) in self._inserted_objects:
            # but it still had to be removed from PostGreSQL, because insert
            # inserted it just before
            del self._inserted_objects[id(obj)]

        self._removed_objects[id(obj)] = obj
        # Just in case the object was modified before removal, let's remove it
        # from the modification list. Note that all sub-objects need to be
        # deleted too!
        for key, reg_obj in list(self._registered_objects.items()):
            if self._get_doc_object(reg_obj) is obj:
                del self._registered_objects[key]
        # We are not doing anything fancy here, since the object might be
        # added again with some different state.

    def setstate(self, obj, doc=None):
        # When reading a state from PostGreSQL, we also need to join the
        # transaction, because we keep an active object cache that gets stale
        # after the transaction is complete and must be cleaned.
        self._join_txn()
        # If the doc is None, but it has been loaded before, we look it
        # up. This acts as a great hook for optimizations that load many
        # documents at once. They can now dump the states into the
        # _latest_states dictionary.
        if doc is None:
            doc = self._latest_states.get(obj._p_oid, None)
        self._reader.set_ghost_state(obj, doc)
        self._loaded_objects[id(obj)] = obj

    def oldstate(self, obj, tid):
        # I cannot find any code using this method. Also, since we do not keep
        # version history, we always raise an error.
        raise KeyError(tid)

    def register(self, obj):
        self._join_txn()

        if self._readonly:
            raise interfaces.ReadOnlyDataManagerError()

        # Do not bring back removed objects. But only main the document
        # objects can be removed, so check for that.
        if id(self._get_doc_object(obj)) in self._removed_objects:
            return

        if obj is not None:
            obj_id = id(obj)
            if obj_id not in self._registered_objects:
                self._registered_objects[obj_id] = obj
                obj_registered = getattr(obj, '_pj_object_registered', None)
                if obj_registered is not None:
                    obj_registered(self)

    def abort(self, transaction):
        if self._conn is None:
            # Connection was never aqcuired - nothing to abort
            return
        self._report_stats()
        try:
            if self._tpc_activated:
                self._conn.tpc_rollback()
            else:
                self._conn.rollback()
        except DISCONNECTED_EXCEPTIONS:
            # this happens usually when PG is restarted and the connection dies
            # our only chance to exit the spiral is to abort the transaction
            pass
        self._release_conn(self._conn)
        self._dirty = False

    def _might_execute_with_error(self, op):
        # run the given method, check for conflicts and DB disconnect
        try:
            op()
        except psycopg2.Error as e:
            check_for_conflict(e, "DataManager.commit")
            check_for_disconnect(e, "DataManager.commit")
            raise

    def setTransactionOptions(
            self,
            readonly: bool = None,
            deferrable: bool = None,
            isolation_level: int = None):
        if isolation_level is not None:
            self._isolation_level = isolation_level
        if readonly is not None:
            self._readonly = readonly
        if deferrable is not None:
            self._deferrable = deferrable

    def _begin(self, transaction):
        self._conn.set_session(isolation_level=self._isolation_level,
                               deferrable=self._deferrable,
                               readonly=self._readonly)

        # This function is called when PJDataManager joins transaction. When
        # two phase commit is requested, we will assign transaction id to
        # underlying connection.
        if not PJ_TWO_PHASE_COMMIT_ENABLED:
            # We don't need to do anything special when two phase commit is
            # disabled. Transaction starts automatically.
            return

        assert self._pristine, ("Error attempting to add data manager "
                                "from old transaction. Create a new "
                                "PJDataManager for new transaction, do not "
                                "modify objects from previously committed "
                                "transactions")
        self._pristine = False

        # Create a global id for the transaction. If it wasn't yet created,
        # create now.
        try:
            txnid = transaction.data(PREPARED_TRANSACTION_ID)
        except KeyError:
            txnid = str(uuid.uuid4())
            transaction.set_data(PREPARED_TRANSACTION_ID, txnid)

        try:
            xid = self._conn.xid(0, txnid, self.database)
            self._conn.tpc_begin(xid)
        except psycopg2.Error as e:
            check_for_disconnect(e, 'PJDataManager._begin')
            raise
        self._tpc_activated = True

    def _log_rw_stats(self):
        if LOG_READ_WRITE_TRANSACTION:
            if self.isDirty():
                LOG.info("PJDataManager transaction had writes")
            else:
                LOG.info("PJDataManager transaction had NO_writes")

    def commit(self, transaction):
        self.flush()
        self._report_stats()
        self._log_rw_stats()

        if not self._tpc_activated:
            try:
                self._might_execute_with_error(self._conn.commit)
                self._dirty = False
            finally:
                self._release_conn(self._conn)

    def tpc_begin(self, transaction):
        pass

    def tpc_vote(self, transaction):
        if self._tpc_activated:
            assert self._conn.status == psycopg2.extensions.STATUS_BEGIN
            if self.isDirty():
                # if the transaction wrote anything we have to call tpc_prepare
                self._might_execute_with_error(self._conn.tpc_prepare)
            else:
                # If the transaction had NO writes

                # We try here hard to avoid calling tpc_prepare when there
                # were NO writes in the current transaction.
                # We found that on AWS Aurora tpc_prepare is SLOW even with no
                # writes.
                if CALL_TPC_PREPARE_ON_NO_WRITE_TRANSACTION:
                    # call tpc_prepare only when the config says so
                    self._might_execute_with_error(self._conn.tpc_prepare)

    def tpc_finish(self, transaction):
        if self._tpc_activated:
            self._report_stats()
            try:
                self._might_execute_with_error(self._conn.tpc_commit)
                self._dirty = False
            finally:
                self._release_conn(self._conn)

    def tpc_abort(self, transaction):
        self.abort(transaction)

    def sortKey(self):
        # `'PJDataManager:%s' % id(self)` would be a better sort key,
        # but this makes our DatamanagerConflictTest lock up.
        return 'PJDataManager:0'

    def _report_stats(self):
        if not PJ_ENABLE_QUERY_STATS:
            return

        stats = self._query_report.calc_and_report()
        TABLE_LOG.info(stats)

    def isDirty(self):
        # this DM is dirty when we had writes or have objects to flush
        # flush on SQL select/with makes this necessary
        #
        # what was written can be checked by looking at TABLE_LOG
        # and self._registered_objects
        return self._dirty or bool(self._registered_objects)

    def setDirty(self):
        if self._readonly:
            raise interfaces.ReadOnlyDataManagerError()
        self._dirty = True

    def __del__(self):
        if self._conn is None:
            return

        LOG.warning("Releasing connection after destroying PJDataManager. "
                    "Active transaction will be aborted.")
        self._release_conn(self._conn)


def get_database_name_from_dsn(dsn):
    import re
    m = re.match(r'.*dbname *= *(.+?)( |$)', dsn)
    if not m:
        LOG.warning("Cannot determine database name from DSN '%s'" % dsn)
        return None

    return m.groups()[0]


def register_query_stats_listener(listener):
    """Register new query stats listener

    All executed sql statements will be reported via the `listner.report()`.
    Note, that queries from all threads will be reported to the same registered
    object.

    `listener` object has to implement `record(sql, args, time, db)` method.
    QueryReport object may be used for this for detailed query analysis.
    """
    GLOBAL_QUERY_STATS_LISTENERS.add(listener)


def unregister_query_stats_listener(listener):
    """Unregister listener, registered by `register_query_stats_listener`
    """
    GLOBAL_QUERY_STATS_LISTENERS.remove(listener)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistence Zope Containers Tests"""
import doctest
import unittest

import mock
import ZODB
import ZODB.DemoStorage
import persistent
import psycopg2
import random
import re
import transaction
import zope.component
import zope.interface
import zope.lifecycleevent
from pprint import pprint
from zope.exceptions import exceptionformatter
from zope.app.testing import placelesssetup
from zope.container.interfaces import IContainer
from zope.container import contained, btree
from zope.interface.verify import verifyObject
from zope.interface.verify import verifyClass
from zope.testing import cleanup, module, renormalizing

from pjpersist import datamanager, interfaces, serialize, testing
from pjpersist.zope import container
from pjpersist.zope import interfaces as zinterfaces

DBNAME = 'pjpersist_container_test'


class ApplicationRoot(container.SimplePJContainer):
    _p_pj_table = 'root'

    def __repr__(self):
        return '<ApplicationRoot>'


class SimplePerson(contained.Contained, persistent.Persistent):
    _p_pj_table = 'person'

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s>' %(self.__class__.__name__, self)


class Person(container.PJContained, SimplePerson):
    pass


def doctest_PJContained_simple():
    """PJContained: simple use

    The simplest way to use PJContained is to use it without any special
    modification. In this case it is required that the container always sets
    the name and parent after loading the item. It can do so directly by
    setting ``_v_name`` and ``_v_parent`` so that the persistence mechanism
    does not kick in.

      >>> class Simples(container.PJContainer):
      ...     def __init__(self, name):
      ...         super(Simples, self).__init__()
      ...         self.name = name
      ...     def __repr__(self):
      ...         return '<Simples %s>' %self.name

      >>> class Simple(container.PJContained, persistent.Persistent):
      ...     pass

    Let's create a simple component and activate the persistence machinery:

      >>> s = Simple()
      >>> s._p_jar = dm

    As you can see, the changed flag is not changed:

      >>> s._p_changed
      False
      >>> s._v_name = 'simple'
      >>> s._v_parent = Simples('one')
      >>> s._p_changed
      False

    And accessing the name and parent works:

      >>> s.__name__
      'simple'
      >>> s.__parent__
      <Simples one>

    But assignment works as well.

      >>> s.__name__ = 'simple2'
      >>> s.__name__
      'simple2'
      >>> s.__parent__ = Simples('two')
      >>> s.__parent__
      <Simples two>
      >>> s._p_changed
      True
    """

def doctest_PJContained_proxy_attr():
    """PJContained: proxy attributes

    It is also possible to use proxy attributes to reference the name and
    parent. This allows you to have nice attribute names for storage in PJ.

    The main benefit, though is the ability of the object to load its
    location, so that you can load the object without going through the
    container and get full location path.

      >>> class Proxies(container.PJContainer):
      ...     def __init__(self, name):
      ...         super(Proxies, self).__init__()
      ...         self.name = name
      ...     def __repr__(self):
      ...         return '<Proxies %s>' %self.name

      >>> class Proxy(container.PJContained, persistent.Persistent):
      ...     _pj_name_attr = 'name'
      ...     _pj_parent_attr = 'parent'
      ...     def __init__(self, name, parent):
      ...         self.name = name
      ...         self.parent = parent

    Let's create a proxy component and activate the persistence machinery:

      >>> p = Proxy('proxy', Proxies('one'))
      >>> p._p_jar = dm

    So accessing the name and parent works:

      >>> p.__name__
      'proxy'
      >>> p.__parent__
      <Proxies one>

    But assignment is only stored into the volatile variables and the proxy
    attribute values are not touched.

      >>> p.__name__ = 'proxy2'
      >>> p.__name__
      'proxy2'
      >>> p.name
      'proxy'
      >>> p.__parent__ = Proxies('two')
      >>> p.__parent__
      <Proxies two>
      >>> p.parent
      <Proxies one>

    This behavior is intentional, so that containment machinery cannot mess
    with the real attributes. Note that in practice, only PJContainer sets
    the ``__name__`` and ``__parent__`` and it should be always consistent
    with the referenced attributes.

    """

def doctest_PJContained_setter_getter():
    """PJContained: setter/getter functions

    If you need ultimate flexibility of where to get and store the name and
    parent, then you can define setters and getters.

      >>> class Funcs(container.PJContainer):
      ...     def __init__(self, name):
      ...         super(Funcs, self).__init__()
      ...         self.name = name
      ...     def __repr__(self):
      ...         return '<Funcs %s>' %self.name

      >>> class Func(container.PJContained, persistent.Persistent):
      ...     _pj_name_getter = lambda s: s.name
      ...     _pj_name_setter = lambda s, v: setattr(s, 'name', v)
      ...     _pj_parent_getter = lambda s: s.parent
      ...     _pj_parent_setter = lambda s, v: setattr(s, 'parent', v)
      ...     def __init__(self, name, parent):
      ...         self.name = name
      ...         self.parent = parent

    Let's create a func component and activate the persistence machinery:

      >>> f = Func('func', Funcs('one'))
      >>> f._p_jar = dm

    So accessing the name and parent works:

      >>> f.__name__
      'func'
      >>> f.__parent__
      <Funcs one>

    In this case, the setters are used, if the name and parent are changed:

      >>> f.__name__ = 'func2'
      >>> f.__name__
      'func2'
      >>> f.name
      'func2'
      >>> f.__parent__ = Funcs('two')
      >>> f.__parent__
      <Funcs two>
      >>> f.parent
      <Funcs two>
    """


def doctest_PJContained_mixed():
    """PJContained: mixed usage

    When the container is stored in the ZODB or another persistence mechanism,
    a mixed usage of proxy attributes and getter/setter functions is the best
    approach.

      >>> class Mixers(btree.BTreeContainer):
      ...     def __init__(self, name):
      ...         super(Mixers, self).__init__()
      ...         self.name = name
      ...     def __repr__(self):
      ...         return '<Mixers %s>' %self.name
      >>> mixers = Mixers('one')

      >>> class Mixer(container.PJContained, persistent.Persistent):
      ...     _pj_name_attr = 'name'
      ...     _pj_parent_getter = lambda s: mixers
      ...     def __init__(self, name):
      ...         self.name = name

    Let's create a mixer component and activate the persistence machinery:

      >>> m = Mixer('mixer')
      >>> m._p_jar = dm

    So accessing the name and parent works:

      >>> m.__name__
      'mixer'
      >>> m.__parent__
      <Mixers one>
    """


def doctest_SimplePJContainer_basic():
    """SimplePJContainer: basic

      >>> cn = 'pjpersist_dot_zope_dot_container_dot_SimplePJContainer'

    Let's make sure events are fired correctly:

      >>> zope.component.provideHandler(handleObjectModifiedEvent)

    Let's add a container to the root:

      >>> dm.root['c'] = container.SimplePJContainer()

      >>> dumpTable(cn)
      [{'data': {'_py_persistent_type': 'pjpersist.zope.container.SimplePJContainer',
                 'data': {}},
        'id': '0001020304050607080a0b0c0'}]

    As you can see, the serialization is very clean. Next we add a person.

      >>> dm.root['c']['stephan'] = SimplePerson('Stephan')
      ContainerModifiedEvent: <...SimplePJContainer ...>
      >>> list(dm.root['c'].keys())
      ['stephan']
      >>> dm.root['c']['stephan']
      <SimplePerson Stephan>

      >>> dm.root['c']['stephan'].__parent__.__class__
      <class 'pjpersist.zope.container.SimplePJContainer'>
      >>> dm.root['c']['stephan'].__name__
      'stephan'

    You can also access objects using the ``get()`` method of course:

      >>> stephan = dm.root['c'].get('stephan')
      >>> stephan.__parent__.__class__
      <class 'pjpersist.zope.container.SimplePJContainer'>
      >>> stephan.__name__
      'stephan'

    Let's commit and access the data again:

      >>> transaction.commit()

      >>> dumpTable('person')
      [{'data': {'__name__': 'stephan',
                 '__parent__': {'_py_type': 'DBREF',
                                 'database': 'pjpersist_test',
                                 'id': '0001020304050607080a0b0c0',
                                 'table': 'pjpersist_dot_zope_dot_container_dot_SimplePJContainer'},
                 '_py_persistent_type': 'pjpersist.zope.tests.test_container.SimplePerson',
                 'name': 'Stephan'},
        'id': '0001020304050607080a0b0c0'}]

      >>> list(dm.root['c'].keys())
      ['stephan']
      >>> dm.root['c']['stephan'].__parent__.__class__
      <class 'pjpersist.zope.container.SimplePJContainer'>
      >>> dm.root['c']['stephan'].__name__
      'stephan'

      >>> dumpTable(cn)
      [{'data': {'_py_persistent_type': 'pjpersist.zope.container.SimplePJContainer',
                 'data': {'stephan': {'_py_type': 'DBREF',
                                        'database': 'pjpersist_test',
                                        'id': '0001020304050607080a0b0c0',
                                        'table': 'person'}}},
        'id': '0001020304050607080a0b0c0'}]

      >>> list(dm.root['c'].items())
      [('stephan', <SimplePerson Stephan>)]

      >>> list(dm.root['c'].values())
      [<SimplePerson Stephan>]

    Now remove the item:

      >>> del dm.root['c']['stephan']
      ContainerModifiedEvent: <...SimplePJContainer ...>

    The changes are immediately visible.

      >>> list(dm.root['c'].keys())
      []
      >>> dm.root['c']['stephan']
      Traceback (most recent call last):
      ...
      KeyError: 'stephan'

    Make sure it is really gone after committing:

      >>> transaction.commit()
      >>> list(dm.root['c'].keys())
      []

    The object is also removed from PJ:

      >>> dumpTable('person')
      []

    Check adding of more objects:

      >>> dm.root['c']['roy'] = SimplePerson('Roy')
      ContainerModifiedEvent: <...SimplePJContainer ...>
      >>> dm.root['c']['adam'] = SimplePerson('Adam')
      ContainerModifiedEvent: <...SimplePJContainer ...>
      >>> dm.root['c']['marius'] = SimplePerson('Marius')
      ContainerModifiedEvent: <...SimplePJContainer ...>

      >>> sorted(dm.root['c'].keys())
      ['adam', 'marius', 'roy']

    """


def doctest_PJContainer_basic():
    """PJContainer: basic

    Let's make sure events are fired correctly:

      >>> zope.component.provideHandler(handleObjectModifiedEvent)

    Let's add a container to the root:

      >>> dm.root['c'] = container.PJContainer('person')

      >>> dumpTable('pjpersist_dot_zope_dot_container_dot_PJContainer')
      [{'data': {'_pj_table': 'person',
                 '_py_persistent_type': 'pjpersist.zope.container.PJContainer'},
        'id': '0001020304050607080a0b0c0'}]

    It is unfortunate that the '_pj_table' attribute is set. This is
    avoidable using a sub-class.

      >>> dm.root['c']['stephan'] = Person('Stephan')
      ContainerModifiedEvent: <...PJContainer ...>
      >>> list(dm.root['c'].keys())
      ['stephan']
      >>> dm.root['c']['stephan']
      <Person Stephan>

      >>> dm.root['c']['stephan'].__parent__.__class__
      <class 'pjpersist.zope.container.PJContainer'>
      >>> dm.root['c']['stephan'].__name__
      'stephan'

    It is a feature of the container that the item is immediately available
    after assignment, but before the data is stored in the database. Let's
    commit and access the data again:

      >>> transaction.commit()

      >>> dumpTable('person')
      [{'data': {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
                 'key': 'stephan',
                 'name': 'Stephan',
                 'parent': {'_py_type': 'DBREF',
                             'database': 'pjpersist_test',
                             'id': '0001020304050607080a0b0c0',
                             'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}},
        'id': '0001020304050607080a0b0c0'}]

      >>> 'stephan' in dm.root['c']
      True
      >>> list(dm.root['c'].keys())
      ['stephan']
      >>> dm.root['c']['stephan'].__parent__.__class__
      <class 'pjpersist.zope.container.PJContainer'>
      >>> dm.root['c']['stephan'].__name__
      'stephan'

    We get a usual key error, if an object does not exist:

      >>> dm.root['c']['roy']
      Traceback (most recent call last):
      ...
      KeyError: 'roy'

      >>> 'roy' in dm.root['c']
      False

    Now remove the item:

      >>> del dm.root['c']['stephan']
      ContainerModifiedEvent: <...PJContainer ...>

    The changes are immediately visible.

      >>> list(dm.root['c'].keys())
      []
      >>> dm.root['c']['stephan']
      Traceback (most recent call last):
      ...
      KeyError: 'stephan'

    Make sure it is really gone after committing:

      >>> transaction.commit()
      >>> list(dm.root['c'].keys())
      []

    Check adding of more objects:

      >>> dm.root['c']['roy'] = SimplePerson('Roy')
      ContainerModifiedEvent: <...PJContainer ...>
      >>> dm.root['c']['adam'] = SimplePerson('Adam')
      ContainerModifiedEvent: <...PJContainer ...>
      >>> dm.root['c']['marius'] = SimplePerson('Marius')
      ContainerModifiedEvent: <...PJContainer ...>

      >>> sorted(dm.root['c'].keys())
      ['adam', 'marius', 'roy']
    """

def doctest_PJContainer_constructor():
    """PJContainer: constructor

    The constructor of the PJContainer class has several advanced arguments
    that allow customizing the storage options.

      >>> c = container.PJContainer(
      ...     'person',
      ...     mapping_key = 'name',
      ...     parent_key = 'site')

      >>> c._pj_mapping_key
      'name'

    The parent key is the key/attribute in which the parent reference is
    stored. This is used to suport multiple containers per PJ table.

      >>> c._pj_parent_key
      'site'
    """

def doctest_PJContainer_pj_parent_key_value():
    r"""PJContainer: _pj_parent_key_value()

    This method is used to extract the parent reference for the item.

      >>> c = container.PJContainer('person')

    The default implementation requires the container to be in some sort of
    persistent store, though it does not care whether this store is PJ or a
    classic ZODB. This feature allows one to mix and match ZODB and PJ
    storage.

      >>> c._pj_get_parent_key_value()
      Traceback (most recent call last):
      ...
      ValueError: _p_jar not found.

    Now the ZODB case:

      >>> c._p_jar = object()
      >>> c._p_oid = b'\x00\x00\x00\x00\x00\x00\x00\x01'
      >>> c._pj_get_parent_key_value()
      'zodb-0000000000000001'

    And finally the PJ case:

      >>> c._p_jar = c._p_oid = None
      >>> dm.root['people'] = c
      >>> c._pj_get_parent_key_value().__class__
      <class 'pjpersist.zope.container.PJContainer'>

    In that final case, the container itself is returned, because upon
    serialization, we simply look up the dbref.
    """

def doctest_PJContainer_many_items():
    """PJContainer: many items

    Let's create an interesting set of data:

      >>> dm.root['people'] = container.PJContainer('person')
      >>> dm.root['people']['stephan'] = Person('Stephan')
      >>> dm.root['people']['roy'] = Person('Roy')
      >>> dm.root['people']['roger'] = Person('Roger')
      >>> dm.root['people']['adam'] = Person('Adam')
      >>> dm.root['people']['albertas'] = Person('Albertas')
      >>> dm.root['people']['russ'] = Person('Russ')

    Let's now search and receive documents as result:

      >>> sorted(dm.root['people'].keys())
      ['adam', 'albertas', 'roger', 'roy', 'russ', 'stephan']
      >>> dm.root['people']['stephan']
      <Person Stephan>
      >>> dm.root['people']['adam']
      <Person Adam>
"""

def doctest_PJContainer_setitem_with_no_key_PJContainer():
    """PJContainer: __setitem__(None, obj)

    Whenever an item is added with no key, getattr(obj, _pj_mapping_key) is used.

      >>> dm.root['people'] = container.PJContainer(
      ...     'person', mapping_key='name')
      >>> dm.root['people'][None] = Person('Stephan')

    Let's now search and receive documents as result:

      >>> sorted(dm.root['people'].keys())
      ['...']
      >>> stephan = list(dm.root['people'].values())[0]
      >>> stephan.__name__ == str(stephan.name)
      True
"""

def doctest_PJContainer_setitem_with_no_key_IdNamesPJContainer():
    """IdNamesPJContainer: __setitem__(None, obj)

    Whenever an item is added with no key, the OID is used.

      >>> dm.root['people'] = container.IdNamesPJContainer('person')
      >>> dm.root['people'][None] = Person('Stephan')

    Let's now search and receive documents as result:

      >>> sorted(dm.root['people'].keys())
      ['...']
      >>> stephan = list(dm.root['people'].values())[0]
      >>> stephan.__name__ == str(stephan._p_oid.id)
      True
"""

def doctest_PJContainer_add_PJContainer():
    """PJContainer: add(value, key=None)

    Sometimes we just do not want to be responsible to determine the name of
    the object to be added. This method makes this optional. The default
    implementation assigns getattr(obj, _pj_mapping_key) as name:

      >>> dm.root['people'] = container.PJContainer(
      ...     'person', mapping_key='name')
      >>> dm.root['people'].add(Person('Stephan'))

    Let's now search and receive documents as result:

      >>> sorted(dm.root['people'].keys())
      ['...']
      >>> stephan = list(dm.root['people'].values())[0]
      >>> stephan.__name__ == str(stephan.name)
      True
"""

def doctest_PJContainer_add_IdNamesPJContainer():
    """IdNamesPJContainer: add(value, key=None)

    Sometimes we just do not want to be responsible to determine the name of
    the object to be added. This method makes this optional. The default
    implementation assigns the OID as name:

      >>> dm.root['people'] = container.IdNamesPJContainer('person')
      >>> dm.root['people'].add(Person('Stephan'))

    Let's now search and receive documents as result:

      >>> sorted(dm.root['people'].keys())
      ['...']
      >>> stephan = list(dm.root['people'].values())[0]
      >>> stephan.__name__ == str(stephan._p_oid.id)
      True
"""

def doctest_PJContainer_concurrent_adds():
    """PJContainer: add(value) can be called from many transactions.

      >>> dm.root['people'] = container.IdNamesPJContainer('person')
      >>> dm.root['people'].add(Person('Roy'))
      >>> commit()

    Let's register a query stats listener:

      >>> class Stats(object):
      ...     def __init__(self):
      ...         self.queries = []
      ...     def record(self, sql, saneargs, *args):
      ...         self.queries.append((sql, saneargs))
      ...
      >>> stats = Stats()
      >>> datamanager.register_query_stats_listener(stats)

    Many persons are added from many connections:

      >>> threads = []
      >>> THREADS = 10
      >>> for i in range(THREADS):
      ...     @testing.run_in_thread
      ...     def add_person():
      ...         conn2 = testing.getConnection(testing.DBNAME)
      ...         try:
      ...             dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))
      ...             ppl = dm2.root['people']
      ...
      ...             ppl.add(Person('Stephan %s' % i))
      ...
      ...             dm2.commit(None)
      ...         finally:
      ...             conn2.close()
      ...     threads.append(add_person)

    Let's wait for threads to finish:

      >>> for thread in threads:
      ...    thread.join()

      >>> datamanager.unregister_query_stats_listener(stats)

    We expect to find THREADS + 1 persons:

      >>> len(dm.root['people'])
      11

    There was one query per person executed (excluding persistence magic):

      >>> len([sql for (sql, args) in stats.queries if 'person' in sql])
      10

    """


def doctest_PJContainer_bool():
  """PJContainers can be evaluated to boolean, however this
  results in an extra query

      >>> dm.root['people'] = container.PJContainer('person')
      >>> bool(dm.root['people'])
      False

      >>> dm.root['people']['stephan'] = Person('Stephan')
      >>> dm.root['people']['roy'] = Person('Roy')

  Enable query statistics and make sure we issue COUNT(*) query instead of
  fetching all the data from the table.

      >>> dm.flush()
      >>> dm._query_report.qlog = []
      >>> with mock.patch.object(datamanager, "PJ_ENABLE_QUERY_STATS", True):
      ...   bool(dm.root['people'])
      True

      >>> dm._query_report.qlog[-1].query
      'SELECT COUNT(person.id) FROM person...'
  """


def doctest_PJContainer_values():
    """PJContainer.values() results in a single query to a database

    We have a container with several items in it

      >>> dm.root['people'] = container.PJContainer('person')
      >>> dm.root['people']['stephan'] = Person('Stephan')
      >>> dm.root['people']['roy'] = Person('Roy')

    To count the queries, enable query statistics


      >>> txn = transaction.manager.get()
      >>> if hasattr(txn, '_v_pj_container_cache'):
      ...     delattr(txn, '_v_pj_container_cache')
      >>> dm._query_report.qlog = []

      >>> with mock.patch.object(datamanager, "PJ_ENABLE_QUERY_STATS", True):
      ...     items = list(dm.root['people'].values())

      >>> from pprint import pprint

    We should have one query to 'persistence_root' table and one query to
    'person` table.'
      >>> len(dm._query_report.qlog)
      2

    """


def doctest_PJContainer_find():
    r"""PJContainer: find

    The PJ Container supports direct PJ queries. It does, however,
    insert the additional container filter arguments and can optionally
    convert the documents to objects.

    Let's create an interesting set of data:

      >>> dm.root['people'] = container.PJContainer('person')
      >>> dm.root['people']['stephan'] = Person('Stephan')
      >>> dm.root['people']['roy'] = Person('Roy')
      >>> dm.root['people']['roger'] = Person('Roger')
      >>> dm.root['people']['adam'] = Person('Adam')
      >>> dm.root['people']['albertas'] = Person('Albertas')
      >>> dm.root['people']['russ'] = Person('Russ')

    Let's now search and receive documents as result:

      >>> import pjpersist.sqlbuilder as sb

      >>> datafld = sb.Field('person', 'data')
      >>> fld = sb.JSON_GETITEM_TEXT(datafld, 'name')
      >>> qry = fld.startswith('Ro')

      # >>> qry.__sqlrepr__('postgres')
      # "(((person.data) ->> ('name')) LIKE ('Ro%') ESCAPE E'\\\\')"

      >>> res = dm.root['people'].raw_find(qry)
      >>> pprint(list(res))
      [['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'roy',
         'name': 'Roy',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'roger',
         'name': 'Roger',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}]]

    And now the same query, but this time with object results:

      >>> res = dm.root['people'].find(qry)
      >>> pprint(list(res))
      [<Person Roy>, <Person Roger>]

    When no spec is specified, all items are returned:

      >>> res = dm.root['people'].find()
      >>> pprint(list(res))
      [<Person Stephan>, <Person Roy>, <Person Roger>, <Person Adam>,
       <Person Albertas>, <Person Russ>]


    Let's see some sqlbuilder parameters (passed along as kwargs):

      >>> res = dm.root['people'].raw_find(orderBy=["(data->'name')"])
      >>> pprint(list(res))
      [['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'adam',
         'name': 'Adam',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'albertas',
         'name': 'Albertas',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'roger',
         'name': 'Roger',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'roy',
         'name': 'Roy',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'russ',
         'name': 'Russ',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}],
       ['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'stephan',
         'name': 'Stephan',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}]]

      >>> res = dm.root['people'].raw_find(orderBy=["(data->'name') DESC"], limit=1)
      >>> pprint(list(res))
      [['0001020304050607080a0b0c0',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
         'key': 'stephan',
         'name': 'Stephan',
         'parent': {'_py_type': 'DBREF',
                     'database': 'pjpersist_test',
                     'id': '0001020304050607080a0b0c0',
                     'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}]]


      >>> res = dm.root['people'].find(orderBy=["(data->'name')"])
      >>> pprint(list(res))
      [<Person Adam>,
       <Person Albertas>,
       <Person Roger>,
       <Person Roy>,
       <Person Russ>,
       <Person Stephan>]

      >>> res = dm.root['people'].find(orderBy=["(data->'name') DESC"], limit=1)
      >>> pprint(list(res))
      [<Person Stephan>]


    You can also search for a single result:

      >>> qry2 = fld.startswith('St')
      >>> res = dm.root['people'].raw_find_one(qry2)
      >>> pprint(res)
      ['0001020304050607080a0b0c0',
       {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
        'key': 'stephan',
        'name': 'Stephan',
        'parent': {'_py_type': 'DBREF',
                    'database': 'pjpersist_test',
                    'id': '0001020304050607080a0b0c0',
                    'table': 'pjpersist_dot_zope_dot_container_dot_PJContainer'}}]

      >>> stephan = dm.root['people'].find_one(qry2)
      >>> pprint(stephan)
      <Person Stephan>

    If no result is found, ``None`` is returned:

      >>> qry3 = fld.startswith('XXX')
      >>> dm.root['people'].find_one(qry3)

    A query or ID must be passed:

      >>> dm.root['people'].find_one()
      Traceback (most recent call last):
      ...
      ValueError: Missing parameter, at least qry or id must be specified.

    On the other hand, if the spec is an id, we look for it instead:

      >>> dm.root['people'].find_one(id=stephan._p_oid.id)
      <Person Stephan>
    """

def doctest_PJ_Container_count():
  """
  count() provides a quick way to count items without fetching them from database

      >>> import pjpersist.sqlbuilder as sb

      >>> dm.root['people'] = container.PJContainer('person')
      >>> int(dm.root['people'].count())
      0

      >>> dm.root['people']['stephan'] = Person('Stephan')
      >>> dm.root['people']['roy'] = Person('Roy')
      >>> dm.root['people']['roger'] = Person('Roger')
      >>> dm.root['people']['adam'] = Person('Adam')
      >>> dm.root['people']['albertas'] = Person('Albertas')
      >>> dm.root['people']['russ'] = Person('Russ')
      >>> int(dm.root['people'].count())
      6

      >>> table = Person._p_pj_table
      >>> datafld = sb.Field('person', 'data')
      >>> fld = sb.JSON_GETITEM_TEXT(datafld, 'name')
      >>> qry = fld.startswith('Ro')
      >>> int(dm.root['people'].count(qry))
      2
  """


def doctest_PJContainer_cache_complete():
    """PJContainer: _cache_complete

    Let's add a bunch of objects:

      >>> ppl = dm.root['people'] = container.PJContainer('person')
      >>> ppl['stephan'] = Person('Stephan')
      >>> ppl['roy'] = Person('Roy')
      >>> ppl['roger'] = Person('Roger')
      >>> ppl['adam'] = Person('Adam')
      >>> ppl['albertas'] = Person('Albertas')
      >>> ppl['russ'] = Person('Russ')

    Clean the cache on the transaction:

      >>> txn = transaction.manager.get()
      >>> if hasattr(txn, '_v_pj_container_cache'):
      ...     delattr(txn, '_v_pj_container_cache')

    The cache is not complete:

      >>> ppl._cache_complete
      False

    We have 6 objects

      >>> len(ppl.items())
      6

    The cache is complete if it's on

      >>> ppl._cache_complete == container.USE_CONTAINER_CACHE
      True

    Del 1

      >>> del ppl['adam']

    5 remain

      >>> len(ppl.items())
      5

    Add 1

      >>> ppl['joe'] = Person('Joe')

    Back to 6

      >>> len(ppl.items())
      6

    The cache is still complete if it's on

      >>> ppl._cache_complete == container.USE_CONTAINER_CACHE
      True

    Clearing the container

      >>> ppl.clear()
      >>> len(ppl.items())
      0

      >>> ppl._cache_complete == container.USE_CONTAINER_CACHE
      True

    """


def doctest_PJContainer_cache_events():
    """PJContainer: _cache insert/delete with events
    (regression: events missed freshly inserted objects)

      >>> ppl = dm.root['people'] = container.PJContainer('person')

    Set cache complete

      >>> ppl.clear()

      >>> ppl._cache_complete
      True

    INSERT
    ------

    Patch event handler:

      >>> @zope.component.adapter(
      ...     zope.interface.Interface,
      ...     zope.lifecycleevent.interfaces.IObjectAddedEvent
      ...     )
      ... def handleObjectAddedEvent(object, event):
      ...     print("container length:", len(ppl))
      ...

      >>> zope.component.provideHandler(handleObjectAddedEvent)

    We want to see that the container in the event handler HAS the just added
    object.

      >>> len(ppl)
      0

    Add a single object

      >>> ppl['stephan'] = Person('Stephan')
      container length: 1

    DELETE
    ------

    Patch event handler:

      >>> @zope.component.adapter(
      ...     zope.interface.Interface,
      ...     zope.lifecycleevent.interfaces.IObjectRemovedEvent
      ...     )
      ... def handleObjectRemovedEvent(object, event):
      ...     print("container length:", len(ppl))
      ...

      >>> zope.component.provideHandler(handleObjectRemovedEvent)

      >>> len(ppl)
      1

    Remove the very first object

      >>> del ppl[list(ppl.keys())[0]]
      container length: 0

    """


def doctest_IdNamesPJContainer_cache_events():
    """IdNamesPJContainer: _cache insert/delete with events
    (regression: events missed freshly inserted objects)

      >>> ppl = dm.root['people'] = container.IdNamesPJContainer('person')

    Set cache complete

      >>> ppl.clear()

      >>> ppl._cache_complete
      True

    INSERT
    ------

    Patch event handler:

      >>> @zope.component.adapter(
      ...     zope.interface.Interface,
      ...     zope.lifecycleevent.interfaces.IObjectAddedEvent
      ...     )
      ... def handleObjectAddedEvent(object, event):
      ...     print("container length:", len(ppl))
      ...

      >>> zope.component.provideHandler(handleObjectAddedEvent)

    We want to see that the container in the event handler HAS the just added
    object.

      >>> len(ppl)
      0

      >>> ppl.add(Person('Stephan'))
      container length: 1

    DELETE
    ------

    Patch event handler:

      >>> @zope.component.adapter(
      ...     zope.interface.Interface,
      ...     zope.lifecycleevent.interfaces.IObjectRemovedEvent
      ...     )
      ... def handleObjectRemovedEvent(object, event):
      ...     print("container length:", len(ppl))
      ...

      >>> zope.component.provideHandler(handleObjectRemovedEvent)

      >>> len(ppl)
      1

    Remove the very first object

      >>> del ppl[list(ppl.keys())[0]]
      container length: 0

    """


def doctest_IdNamesPJContainer_basic():
    """IdNamesPJContainer: basic

    This container uses the PJ ObjectId as the name for each object. Since
    ObjectIds are required to be unique within a table, this is actually
    a nice and cheap scenario.

    Let's add a container to the root:

      >>> dm.root['c'] = container.IdNamesPJContainer('person')

    Let's now add a new person:

      >>> dm.root['c'].add(Person('Stephan'))
      >>> keys = list(dm.root['c'].keys())
      >>> keys
      ['0001020304050607080a0b0c0']
      >>> name = keys[0]
      >>> dm.root['c'][name]
      <Person Stephan>

      >>> list(dm.root['c'].values())
      [<Person Stephan>]

      >>> dm.root['c'][name].__parent__.__class__
      <class 'pjpersist.zope.container.IdNamesPJContainer'>
      >>> dm.root['c'][name].__name__
      '0001020304050607080a0b0c0'

    It is a feature of the container that the item is immediately available
    after assignment, but before the data is stored in the database. Let's
    commit and access the data again:

      >>> transaction.commit()

      >>> dumpTable('person')
      [{'data': {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
                 'name': 'Stephan',
                 'parent': {'_py_type': 'DBREF',
                             'database': 'pjpersist_test',
                             'id': '0001020304050607080a0b0c0',
                             'table': 'pjpersist_dot_zope_dot_container_dot_IdNamesPJContainer'}},
        'id': '0001020304050607080a0b0c0'}]

    Notice how there is no "key" entry in the document. We get a usual key
    error, if an object does not exist:

      >>> dm.root['c']['0001020304050607080a0b0c0']
      Traceback (most recent call last):
      ...
      KeyError: '0001020304050607080a0b0c0'

      >>> '0001020304050607080a0b0c0' in dm.root['c']
      False

      >>> dm.root['c']['roy']
      Traceback (most recent call last):
      ...
      KeyError: 'roy'

      >>> 'roy' in dm.root['c']
      False

    Now remove the item:

      >>> dm.root['c'][name]
      <Person Stephan>

      >>> del dm.root['c'][name]

    The changes are immediately visible.

      >>> list(dm.root['c'].keys())
      []
      >>> dm.root['c'][name]
      Traceback (most recent call last):
      ...
      KeyError: '0001020304050607080a0b0c0'

    Make sure it is really gone after committing:

      >>> transaction.commit()
      >>> list(dm.root['c'].keys())
      []
    """

def doctest_load_one_ignore_cache():
    """PJContainer._load_one() can be instructed to ignore the cache.

    Let's add some objects:

      >>> dm.root['people'] = people = People()
      >>> people[None] = roy = Person('Roy')

      >>> people[roy.__name__] == roy
      True

      >>> people._load_one(roy.__name__, {'name': 'Roy'}) is roy
      True

      >>> stephan = people._load_one(
      ...     roy.__name__,
      ...     {'_py_persistent_type':
      ...          'pjpersist.zope.tests.test_container.Person',
      ...      'name': 'Stephan'},
      ...     use_cache=False)
      >>> stephan.name
      'Stephan'
    """

def doctest_AllItemsPJContainer_basic():
    """AllItemsPJContainer: basic

    This type of container returns all items of the table without regard
    of a parenting hierarchy.

    Let's start by creating two person containers that service different
    purposes:

      >>> dm.root['friends'] = container.PJContainer('person')
      >>> dm.root['friends']['roy'] = Person('Roy')
      >>> dm.root['friends']['roger'] = Person('Roger')

      >>> dm.root['family'] = container.PJContainer('person')
      >>> dm.root['family']['anton'] = Person('Anton')
      >>> dm.root['family']['konrad'] = Person('Konrad')

      >>> transaction.commit()
      >>> sorted(dm.root['friends'].keys())
      ['roger', 'roy']
      >>> sorted(dm.root['family'].keys())
      ['anton', 'konrad']

    Now we can create an all-items-container that allows us to view all
    people.

      >>> dm.root['all-people'] = container.AllItemsPJContainer('person')
      >>> sorted(dm.root['all-people'].keys())
      ['anton', 'konrad', 'roger', 'roy']
    """

def doctest_SubDocumentPJContainer_basic():
    r"""SubDocumentPJContainer: basic

    Let's make sure events are fired correctly:

      >>> zope.component.provideHandler(handleObjectModifiedEvent)

    Sub_document PJ containers are useful, since they avoid the creation of
    a commonly trivial tables holding meta-data for the table
    object. But they require a root document:

      >>> dm.root['app_root'] = ApplicationRoot()

    Let's add a container to the app root:

      >>> dm.root['app_root']['people'] = \
      ...     container.SubDocumentPJContainer('person')
      ContainerModifiedEvent: <ApplicationRoot>

      >>> transaction.commit()
      >>> dumpTable('root')
      [{'data':
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.ApplicationRoot',
         'data':
           {'people': {'_pj_table': 'person',
                        '_py_persistent_type': 'pjpersist.zope.container.SubDocumentPJContainer'}}},
        'id': '0001020304050607080a0b0c0'}]

    It is unfortunate that the '_pj_table' attribute is set. This is
    avoidable using a sub-class. Let's make sure the container can be loaded
    correctly:

      >>> dm.root['app_root']['people'].__class__
      <class 'pjpersist.zope.container.SubDocumentPJContainer'>
      >>> dm.root['app_root']['people'].__parent__
      <ApplicationRoot>
      >>> dm.root['app_root']['people'].__name__
      'people'

    Let's add an item to the container:

      >>> dm.root['app_root']['people']['stephan'] = Person('Stephan')
      ContainerModifiedEvent: <...SubDocumentPJContainer ...>
      >>> list(dm.root['app_root']['people'].keys())
      ['stephan']
      >>> dm.root['app_root']['people']['stephan']
      <Person Stephan>

      >>> transaction.commit()
      >>> list(dm.root['app_root']['people'].keys())
      ['stephan']
    """

def doctest_PJContainer_with_ZODB():
    r"""PJContainer: with ZODB

    This test demonstrates how a PJ Container lives inside a ZODB tree:

      >>> zodb = ZODB.DB(ZODB.DemoStorage.DemoStorage())
      >>> root = zodb.open().root()
      >>> root['app'] = btree.BTreeContainer()
      >>> root['app']['people'] = container.PJContainer('person')

    Let's now commit the transaction and make sure everything is cool.

      >>> transaction.commit()
      >>> root = zodb.open().root()
      >>> root['app'].__class__
      <class 'zope.container.btree.BTreeContainer'>
      >>> root['app']['people'].__class__
      <class 'pjpersist.zope.container.PJContainer'>

    So let's try again:

      >>> list(root['app']['people'].keys())
      []

    Next we create a person object and make sure it gets properly persisted.

      >>> root['app']['people']['stephan'] = Person('Stephan')
      >>> transaction.commit()
      >>> root = zodb.open().root()
      >>> list(root['app']['people'].keys())
      ['stephan']

      >>> stephan = root['app']['people']['stephan']
      >>> stephan.__name__
      'stephan'
      >>> stephan.__parent__.__class__
      <class 'pjpersist.zope.container.PJContainer'>

      >>> dumpTable('person')
      [{'data': {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Person',
                 'key': 'stephan',
                 'name': 'Stephan',
                 'parent': 'zodb-01af3b00c5'},
        'id': '0001020304050607080a0b0c0'}]

    Note that we produced a nice hex-presentation of the ZODB's OID.
    """


# classes for doctest_Realworldish
class Campaigns(container.PJContainer):
    _pj_table = 'campaigns'

    def __init__(self, name):
        self.name = name
        super(Campaigns, self).__init__()

    def add(self, campaign):
        self[campaign.name] = campaign

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)


class PJItem(container.PJContained,
                persistent.Persistent):
    pass


class Campaign(PJItem, container.PJContainer):
    _pj_table = 'persons'
    _p_pj_table = 'campaigns'

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)


class PersonItem(PJItem):
    _p_pj_table = 'persons'

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)


def doctest_Realworldish():
    """Let's see some real worldish hierarchic structure is persisted

    Let's make sure events are fired correctly:

      >>> zope.component.provideHandler(handleObjectModifiedEvent)

    Let's add a container to the root:

      >>> dm.root['c'] = Campaigns('foobar')

      >>> dumpTable(
      ...     'pjpersist_dot_zope_dot_tests_dot_test_container_dot_Campaigns')
      [{'data': {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Campaigns',
                 'name': 'foobar'},
        'id': '0001020304050607080a0b0c0'}]

    It is unfortunate that the '_pj_table' attribute is set. This is
    avoidable using a sub-class.

      >>> dm.root['c']['one'] = Campaign('one')
      ContainerModifiedEvent: <...Campaigns ...>
      >>> list(dm.root['c'].keys())
      ['one']
      >>> dm.root['c']['one']
      <Campaign one>

      >>> dm.root['c']['one'].__parent__
      <Campaigns foobar>
      >>> dm.root['c']['one'].__name__
      'one'

    It is a feature of the container that the item is immediately available
    after assignment, but before the data is stored in the database. Let's
    commit and access the data again:

      >>> transaction.commit()

      >>> dumpTable(Campaigns._pj_table)
      [{'data': {'_py_persistent_type': 'pjpersist.zope.tests.test_container.Campaign',
                 'key': 'one',
                 'name': 'one',
                 'parent': {'_py_type': 'DBREF',
                             'database': 'pjpersist_test',
                             'id': '0001020304050607080a0b0c0',
                             'table': 'pjpersist_dot_zope_dot_tests_dot_test_container_dot_Campaigns'}},
        'id': '0001020304050607080a0b0c0'}]

      >>> 'one' in dm.root['c']
      True
      >>> list(dm.root['c'].keys())
      ['one']
      >>> dm.root['c']['one'].__parent__
      <Campaigns foobar>
      >>> dm.root['c']['one'].__name__
      'one'

    We get a usual key error, if an object does not exist:

      >>> dm.root['c']['roy']
      Traceback (most recent call last):
      ...
      KeyError: 'roy'

      >>> 'roy' in dm.root['c']
      False

    Now remove the item:

      >>> del dm.root['c']['one']
      ContainerModifiedEvent: <...Campaigns ...>

    The changes are immediately visible.

      >>> list(dm.root['c'].keys())
      []
      >>> dm.root['c']['one']
      Traceback (most recent call last):
      ...
      KeyError: 'one'

    Make sure it is really gone after committing:

      >>> transaction.commit()
      >>> list(dm.root['c'].keys())
      []

    Check adding of more objects:

      >>> dm.root['c']['1'] = c1 = Campaign('One')
      ContainerModifiedEvent: <...Campaigns ...>
      >>> dm.root['c']['2'] = c2 = Campaign('Two')
      ContainerModifiedEvent: <...Campaigns ...>
      >>> dm.root['c']['3'] = Campaign('Three')
      ContainerModifiedEvent: <...Campaigns ...>

      >>> sorted(dm.root['c'].keys())
      ['1', '2', '3']

    Check adding of more subitems:

      >>> stephan = c1['stephan'] = PersonItem('Stephan')
      ContainerModifiedEvent: <Campaign One>
      >>> roy = c1['roy'] = PersonItem('Roy')
      ContainerModifiedEvent: <Campaign One>

      >>> sorted(c1.keys())
      ['roy', 'stephan']

      >>> adam = c2['adam'] = PersonItem('Adam')
      ContainerModifiedEvent: <Campaign Two>

      >>> sorted(c1.keys())
      ['roy', 'stephan']
      >>> sorted(c2.keys())
      ['adam']

    """


class People(container.AllItemsPJContainer):
    _pj_mapping_key = 'name'
    _p_pj_table = 'people'
    _pj_table = 'person'


class Address(persistent.Persistent):
    _p_pj_table = 'address'

    def __init__(self, city):
        self.city = city


class PeoplePerson(persistent.Persistent, container.PJContained):
    _p_pj_table = 'person'

    def __init__(self, name, age):
        self.name = name
        self.age = age
        self.address = Address('Boston %i' %age)

    def __repr__(self):
        return '<%s %s @ %i [%s]>' %(
            self.__class__.__name__, self.name, self.age, self.__name__)


def doctest_load_does_not_set_p_changed():
    """We need to guarantee that _p_changed is not True on obj load

    Let's add some objects:

      >>> dm.root['people'] = people = People()
      >>> for idx in range(2):
      ...     people[None] = PeoplePerson('Mr Number %.5i' %idx, random.randint(0, 100))
      >>> transaction.commit()

      >>> objs = [o for o in people.values()]
      >>> len(objs)
      2
      >>> [o._p_changed for o in objs]
      [False, False]

      >>> [o._p_changed for o in people.values()]
      [False, False]

      >>> transaction.commit()

      >>> x = transaction.begin()
      >>> [o._p_changed for o in people.values()]
      [False, False]

      >>> [o._p_changed for o in people.values()]
      [False, False]

    """


def doctest_firing_events_PJContainer():
    """Events need to be fired when _pj_mapping_key is already set on the object
    and the object gets added to the container

      >>> @zope.component.adapter(zope.interface.interfaces.IObjectEvent)
      ... def eventHandler(event):
      ...     print(event)

      >>> zope.component.provideHandler(eventHandler)

    Let's add some objects:

      >>> dm.root['people'] = people = People()
      >>> for idx in range(2):
      ...     people[None] = PeoplePerson('Mr Number %.5i' %idx, random.randint(0, 100))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['Mr Number 00000', 'Mr Number 00001']

      >>> for idx in range(2):
      ...     name = 'Mr Number %.5i' % (idx+10, )
      ...     people.add(PeoplePerson(name, random.randint(0, 100)))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['Mr Number 00000', 'Mr Number 00001', 'Mr Number 00010', 'Mr Number 00011']

      >>> for idx in range(2):
      ...     name = 'Mr Number %.5i' % (idx+20, )
      ...     people[name] = PeoplePerson(name, random.randint(0, 100))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['Mr Number 00000', 'Mr Number 00001', 'Mr Number 00010', 'Mr Number 00011',
       'Mr Number 00020', 'Mr Number 00021']

    """


class PeopleWithIDKeys(container.IdNamesPJContainer):
    _p_pj_table = 'people'
    _pj_table = 'person'


def doctest_firing_events_IdNamesPJContainer():
    """Events need to be fired when the object gets added to the container

      >>> @zope.component.adapter(zope.interface.interfaces.IObjectEvent)
      ... def eventHandler(event):
      ...     print(event)

      >>> zope.component.provideHandler(eventHandler)

    Let's add some objects:

      >>> dm.root['people'] = people = PeopleWithIDKeys()
      >>> for idx in range(2):
      ...     people[None] = PeoplePerson('Mr Number %.5i' %idx, random.randint(0, 100))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['4e7ddf12e138237403000000', '4e7ddf12e138237403000000']

      >>> for idx in range(2):
      ...     name = 'Mr Number %.5i' % (idx+10, )
      ...     people.add(PeoplePerson(name, random.randint(0, 100)))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['4e7ddf12e138237403000000', '4e7ddf12e138237403000000', '4e7ddf12e138237403000000', '4e7ddf12e138237403000000']

    We can set custom keys as well, they will end up in mongo documents as _id
    attributes.

      >>> for idx in range(2):
      ...     name = '4e7ddf12e1382374030%.5i' % (idx+20, )
      ...     people[name] = PeoplePerson(name, random.randint(0, 100))
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>
      <zope.lifecycleevent.ObjectAddedEvent object at ...>
      <zope.container.contained.ContainerModifiedEvent object at ...>

      >>> transaction.commit()
      >>> list(people.keys())
      ['4e7ddf12e138237403000000', '4e7ddf12e138237403000000',
      '4e7ddf12e138237403000000', '4e7ddf12e138237403000000',
      '4e7ddf12e138237403000000', '4e7ddf12e138237403000000']

    """

import zope.schema
from pjpersist.persistent import SimpleColumnSerialization, select_fields


class IPerson(zope.interface.Interface):
    name = zope.schema.TextLine(title='Name')
    address = zope.schema.TextLine(title='Address')
    visited = zope.schema.Datetime(title='Visited')
    phone = zope.schema.TextLine(title='Phone')


class ColumnPeople(container.AllItemsPJContainer):
    _pj_mapping_key = 'name'
    _p_pj_table = 'people'
    _pj_table = 'cperson'

    _pj_column_fields = container.AllItemsPJContainer._pj_column_fields + (
        'name',)


@zope.interface.implementer(IPerson)
class ColumnPerson(SimpleColumnSerialization, container.PJContained,
                   persistent.Persistent):
    _p_pj_table = 'cperson'

    _pj_column_fields = select_fields(IPerson, 'name')

    def __init__(self, name, phone=None, address=None, friends=None,
                 visited=(), birthday=None):
        self.name = name
        self.address = address
        self.friends = friends or {}
        self.visited = visited
        self.phone = phone
        self.birthday = birthday

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)


def doctest_PJContainer_SimpleColumnSerialization():
    """
    Check PJContainer methods when SimpleColumnSerialization is in effect

      >>> import pjpersist.sqlbuilder as sb

      >>> table = ColumnPerson._p_pj_table
      >>> dm._ensure_sql_columns(ColumnPerson('foo'), table)

      >>> dm.root['people'] = people = ColumnPeople()
      >>> for idx in range(20):
      ...     people[None] = ColumnPerson('Mr Number %.5i' %idx)

      >> dumpTable('cperson')

      >>> qry = sb.Field(table, 'name') == 'Mr Number 00010'
      >>> list(people.find(qry))
      [<ColumnPerson Mr Number 00010>]

      >>> pprint(list(people.raw_find(qry)))
      [['54894d3fb25d2b232e0046d6',
        'Mr Number 00010',
        {'_py_persistent_type': 'pjpersist.zope.tests.test_container.ColumnPerson',
         'address': None,
         'birthday': None,
         'friends': {},
         'name': 'Mr Number 00010',
         'phone': None,
         'visited': []}]]

      >>> people.find_one(qry)
      <ColumnPerson Mr Number 00010>

      >>> pprint(people.raw_find_one(qry))
      ['54894d80b25d2b240f00bbf6',
       'Mr Number 00010',
       {'_py_persistent_type': 'pjpersist.zope.tests.test_container.ColumnPerson',
        'address': None,
        'birthday': None,
        'friends': {},
        'name': 'Mr Number 00010',
        'phone': None,
        'visited': []}]

      >>> people['Mr Number 00007']
      <ColumnPerson Mr Number 00007>

      >>> del people['Mr Number 00007']
      >>> people['Mr Number 00007']
      Traceback (most recent call last):
      ...
      KeyError: 'Mr Number 00007'

      >>> pprint(list(people))
      ['Mr Number 00000',
       'Mr Number 00001',
       'Mr Number 00002',
       'Mr Number 00003',
       'Mr Number 00004',
       'Mr Number 00005',
       'Mr Number 00006',
       'Mr Number 00008',
       'Mr Number 00009',
       'Mr Number 00010',
       'Mr Number 00011',
       'Mr Number 00012',
       'Mr Number 00013',
       'Mr Number 00014',
       'Mr Number 00015',
       'Mr Number 00016',
       'Mr Number 00017',
       'Mr Number 00018',
       'Mr Number 00019']

      >>> pprint(list(people.keys()))
      ['Mr Number 00000',
       'Mr Number 00001',
       'Mr Number 00002',
       'Mr Number 00003',
       'Mr Number 00004',
       'Mr Number 00005',
       'Mr Number 00006',
       'Mr Number 00008',
       'Mr Number 00009',
       'Mr Number 00010',
       'Mr Number 00011',
       'Mr Number 00012',
       'Mr Number 00013',
       'Mr Number 00014',
       'Mr Number 00015',
       'Mr Number 00016',
       'Mr Number 00017',
       'Mr Number 00018',
       'Mr Number 00019']
    """


def doctest_PJContainer_get_sb_fields():
    """Check _get_sb_fields behavour

      >>> c = ColumnPeople()

      >>> c._get_sb_fields(())
      cperson.*

      >>> c._get_sb_fields(None)
      cperson.*

      >>> def printit(res):
      ...     for r in res:
      ...           print(r.__sqlrepr__('postgres'))

    Native column

      >>> printit(c._get_sb_fields(('name',)))
      cperson.name AS name

    Simple first level JSONB field

      >>> printit(c._get_sb_fields(('visited',)))
      ((cperson.data) -> ('visited')) AS visited

    A combination

      >>> printit(c._get_sb_fields(('id', 'visited',)))
      cperson.id AS id
      ((cperson.data) -> ('visited')) AS visited

    Supports any depth of a JSONB structure

      >>> printit(c._get_sb_fields(('id', 'metadata.number')))
      cperson.id AS id
      ((cperson.data) #> (array['metadata', 'number'])) AS metadata_number

    """


class ContainerConflictTest(testing.PJTestCase):

    def test_conflict_SimplePJContainer_key_1(self):
        """Check conflict detection. We add the same key in
        different transactions, simulating separate processes."""

        self.dm.root['c'] = container.SimplePJContainer()
        transaction.commit()

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        dm2.root['c']['stephan'] = SimplePerson('Stephan')

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))

        dm1.root['c']['stephan'] = SimplePerson('Stephan')

        #Finish in order 2 - 1

        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_SimplePJContainer_key_2(self):
        """Check conflict detection. We add the same key in
        different transactions, simulating separate processes."""

        self.dm.root['c'] = container.SimplePJContainer()
        transaction.commit()

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))

        dm1.root['c']['stephan'] = SimplePerson('Stephan')

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        dm2.root['c']['stephan'] = SimplePerson('Stephan')

        #Finish in order 1 - 2
        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_PJContainer_key_1(self):
        """Check conflict detection. We add the same key in
        different transactions, simulating separate processes."""

        self.dm.root['c'] = container.PJContainer('person')
        # (auto-create the table)
        self.dm.root['c']['roy'] = Person('Roy')
        transaction.commit()

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        dm2.root['c']['stephan'] = Person('Stephan')

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))

        dm1.root['c']['stephan'] = Person('Stephan')

        #Finish in order 2 - 1
        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_PJContainer_key_2(self):
        """Check conflict detection. We add the same key in
        different transactions, simulating separate processes."""

        self.dm.root['c'] = container.PJContainer('person')
        # (auto-create the table)
        self.dm.root['c']['roy'] = Person('Roy')
        transaction.commit()

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))

        dm1.root['c']['stephan'] = Person('Stephan')

        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))

        dm2.root['c']['stephan'] = Person('Stephan')

        #Finish in order 1 - 2
        with self.assertRaises(interfaces.ConflictError):
            transaction.commit()

        transaction.abort()

        conn2.close()
        conn1.close()

    def test_conflict_PJContainer_same_id(self):
        """Check conflict detection. Should never happen that
        PJDataManager.createId creates the very same ID, but you never know..."""

        self.dm.root['c'] = container.PJContainer('person')
        # (auto-create the table)
        self.dm.root['c']['roy'] = Person('Roy')
        transaction.commit()

        conn1 = testing.getConnection(testing.DBNAME)
        dm1 = datamanager.PJDataManager(testing.DummyConnectionPool(conn1))
        dm1.createId = lambda: 'abcd'
        conn2 = testing.getConnection(testing.DBNAME)
        dm2 = datamanager.PJDataManager(testing.DummyConnectionPool(conn2))
        dm2.createId = lambda: 'abcd'

        dm1.root['c']['stephan1'] = Person('Stephan1')

        # pain: isolation would block inserting stephan2
        #       we have to commit dm1 first
        dm1.commit(None)

        with self.assertRaises(psycopg2.IntegrityError):
            # XXX: this might need to be translated to ConflictError
            dm2.root['c']['stephan2'] = Person('Stephan2')

        transaction.abort()

        conn2.close()
        conn1.close()


class PJContainedInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        from zope.location.interfaces import IContained
        self.assertTrue(verifyClass(IContained, container.PJContained))

    def test_verifyObject(self):
        from zope.location.interfaces import IContained
        self.assertTrue(verifyObject(IContained, container.PJContained()))


class SimplePJContainerInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        self.assertTrue(verifyClass(IContainer, container.SimplePJContainer))

    def test_verifyObject(self):
        self.assertTrue(verifyObject(IContainer, container.SimplePJContainer()))


class PJContainerInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        self.assertTrue(
            verifyClass(IContainer, container.PJContainer))
        self.assertTrue(
            verifyClass(zinterfaces.IPJContainer, container.PJContainer))

    def test_verifyObject(self):
        self.assertTrue(
            verifyObject(IContainer, container.PJContainer()))
        self.assertTrue(
            verifyObject(zinterfaces.IPJContainer, container.PJContainer()))


class IdNamesPJContainerInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        self.assertTrue(
            verifyClass(IContainer, container.IdNamesPJContainer))
        self.assertTrue(
            verifyClass(zinterfaces.IPJContainer, container.IdNamesPJContainer))

    def test_verifyObject(self):
        self.assertTrue(
            verifyObject(IContainer, container.IdNamesPJContainer()))
        self.assertTrue(
            verifyObject(zinterfaces.IPJContainer, container.IdNamesPJContainer()))


class AllItemsPJContainerInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        self.assertTrue(
            verifyClass(IContainer, container.AllItemsPJContainer))
        self.assertTrue(
            verifyClass(zinterfaces.IPJContainer, container.AllItemsPJContainer))

    def test_verifyObject(self):
        self.assertTrue(
            verifyObject(IContainer, container.AllItemsPJContainer()))
        self.assertTrue(
            verifyObject(zinterfaces.IPJContainer, container.AllItemsPJContainer()))


class SubDocumentPJContainerInterfaceTest(unittest.TestCase):

    def test_verifyClass(self):
        self.assertTrue(
            verifyClass(IContainer, container.SubDocumentPJContainer))
        self.assertTrue(
            verifyClass(zinterfaces.IPJContainer, container.SubDocumentPJContainer))

    def test_verifyObject(self):
        self.assertTrue(
            verifyObject(IContainer, container.SubDocumentPJContainer()))
        self.assertTrue(
            verifyObject(zinterfaces.IPJContainer, container.SubDocumentPJContainer()))


checker = renormalizing.RENormalizing([
    (re.compile(r'datetime.datetime(.*)'),
     'datetime.datetime(2011, 10, 1, 9, 45)'),
    (re.compile(r"'[0-9a-f]{24}'"),
     "'0001020304050607080a0b0c0'"),
    (re.compile(r"object at 0x[0-9a-f]*>"),
     "object at 0x001122>"),
    (re.compile(r"'zodb-[0-9a-f]+'"),
     "'zodb-01af3b00c5'"),
    (re.compile(r"zodb-[0-9a-f]+"),
     "zodb-01af3b00c5"),
    # Mangle unicode strings
    (re.compile("u('.*?')"), r"\1"),
    (re.compile('u(".*?")'), r"\1"),
    ])

@zope.component.adapter(
    zope.interface.Interface,
    zope.lifecycleevent.interfaces.IObjectModifiedEvent
    )
def handleObjectModifiedEvent(object, event):
    print(event.__class__.__name__+':', repr(object))


def setUp(test):
    placelesssetup.setUp(test)
    testing.setUp(test)

    # since the table gets created in PJContainer.__init__ we need to provide
    # a IPJDataManagerProvider
    @zope.interface.implementer(interfaces.IPJDataManagerProvider)
    class Provider(object):

        def get(self, database):
            return test.globs['dm']

    zope.component.provideUtility(Provider())

    # silence this, otherwise half-baked objects raise exceptions
    # on trying to __repr__ missing attributes
    test.orig_DEBUG_EXCEPTION_FORMATTER = \
        exceptionformatter.DEBUG_EXCEPTION_FORMATTER
    exceptionformatter.DEBUG_EXCEPTION_FORMATTER = 0

def noCacheSetUp(test):
    container.USE_CONTAINER_CACHE = False
    setUp(test)

def tearDown(test):
    testing.tearDown(test)
    placelesssetup.tearDown(test)
    exceptionformatter.DEBUG_EXCEPTION_FORMATTER = \
        test.orig_DEBUG_EXCEPTION_FORMATTER

def noCacheTearDown(test):
    container.USE_CONTAINER_CACHE = True
    tearDown(test)

def test_suite():
    dt = doctest.DocTestSuite(
                setUp=setUp, tearDown=tearDown, checker=checker,
                optionflags=(doctest.NORMALIZE_WHITESPACE|
                             doctest.ELLIPSIS|
                             doctest.REPORT_ONLY_FIRST_FAILURE
                             #|doctest.REPORT_NDIFF
                             ),
                )
    dt.layer = testing.db_layer

    return unittest.TestSuite((
        dt,
        #doctest.DocTestSuite(
        #        setUp=noCacheSetUp, tearDown=noCacheTearDown, checker=checker,
        #        optionflags=(doctest.NORMALIZE_WHITESPACE|
        #                     doctest.ELLIPSIS|
        #                     doctest.REPORT_ONLY_FIRST_FAILURE
        #                     #|doctest.REPORT_NDIFF
        #                     )
        #        ),
        unittest.makeSuite(ContainerConflictTest),
        unittest.makeSuite(PJContainedInterfaceTest),
        unittest.makeSuite(SimplePJContainerInterfaceTest),
        unittest.makeSuite(PJContainerInterfaceTest),
        unittest.makeSuite(IdNamesPJContainerInterfaceTest),
        unittest.makeSuite(AllItemsPJContainerInterfaceTest),
        unittest.makeSuite(SubDocumentPJContainerInterfaceTest),
        ))



##############################################################################
#
# Copyright (c) 2013 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Annotations Implementation."""
from persistent.dict import PersistentDict
from zope import component, interface
from zope.annotation import interfaces

from collections import MutableMapping

class IPJAttributeAnnotatable(interfaces.IAnnotatable):
    """Marker indicating that annotations can be stored on an attribute.

    This is a marker interface giving permission for an `IAnnotations`
    adapter to store data in an attribute named `__annotations__`.

    """

def normalize_key(key):
    return key.replace('.', '_')

@interface.implementer(interfaces.IAnnotations)
@component.adapter(IPJAttributeAnnotatable)
class AttributeAnnotations(MutableMapping):
    """Store annotations on an object

    Store annotations in the `__annotations__` attribute on a
    `IAttributeAnnotatable` object.
    """

    def __init__(self, obj, context=None):
        self.obj = obj

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    def get(self, key, default=None):
        """See zope.annotation.interfaces.IAnnotations"""
        key = normalize_key(key)
        return getattr(self.obj, key, default)

    def __getitem__(self, key):
        key = normalize_key(key)
        try:
            return getattr(self.obj, key)
        except AttributeError:
            raise KeyError(key)

    def keys(self):
        annotations = getattr(self.obj, self.ATTR_NAME, None)
        if annotations is None:
            return []

        return annotations.keys()

    def __iter__(self):
        annotations = getattr(self.obj, self.ATTR_NAME, None)
        if annotations is None:
            return iter([])

        return iter(annotations)

    def __len__(self):
        raise NotImplementedError

    def __setitem__(self, key, value):
        """See zope.annotation.interfaces.IAnnotations"""
        key = normalize_key(key)
        setattr(self.obj, key, value)

    def __delitem__(self, key):
        """See zope.app.interfaces.annotation.IAnnotations"""
        key = normalize_key(key)
        try:
            delattr(self.obj, key)
        except AttributeError:
            raise KeyError(key)



##############################################################################
#
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""SQLBuilder extensions"""
import json, re

from sqlobject.sqlbuilder import *


########################################
## Postgres JSON operators
########################################

# Let's have them here and then see if they're worth contributing
# upstream when they settle down.

class PGArray(SQLExpression):
    """PostgreSQL array expression.

    The downside of current implementation is that the items must be
    the same type.  Literal arrays like '{1, 1.5, "two"}' can be
    non-homogenous, but need different quoting of strings.
    """

    def __init__(self, iterable):
        self.iterable = iterable

    def __sqlrepr__(self, db):
        assert db == 'postgres', "Postgres-specific feature, sorry."
        items = (sqlrepr(item, db) for item in self.iterable)
        return 'array[%s]' % ", ".join(items)


class PGArrayLiteral(SQLExpression):
    """PostgreSQL array literal expression.

    This works with all types except floats because paths are
    separated by periods also.  This is mostly used for dict/array keys so
    it shouldn't be much of a problem
    Literal arrays don't need quotes around the values like '{a, b, 1, 2}',
    but optional double quoting of string values is valid and helpful certain
    cases, like words with apostrophes I.E '{"Bob''s"}'
    """

    def __init__(self, iterable):
        self.iterable = iterable

    def __sqlrepr__(self, db):
        assert db == 'postgres', "Postgres-specific feature, sorry."
        # strip quotes off of nested array literals
        items = (sqlrepr(item, db) if not
                 isinstance(item, self.__class__) else sqlrepr(item, db)[1:-1]
                 for item in self.iterable)
        literal = '{%s}' % ", ".join(items)
        # switches single quotes to double quotes if there aren't
        # 2 single quotes in a row, which indicates an escaped apostrophe
        regex = re.compile("(?<!')'(?!')")
        return "'{}'".format(regex.sub('"', literal))


class TYPECAST(SQLExpression):
    """Cast to a type"""

    cast = None

    def __init__(self, arg, typ=None):
        self.arg = arg
        if typ is not None:
            self.cast = '::' + typ

    def __sqlrepr__(self, db):
        return sqlrepr(self.arg, db) + self.cast


class JSONB(TYPECAST):
    """Cast to JSONB"""
    cast = "::jsonb"


class JSON(TYPECAST):
    """Cast to JSON"""
    cast = '::json'


class TEXT(TYPECAST):
    """Cast to text"""
    cast = '::text'


def JSON_GETITEM(json, key):
    return SQLOp("->", json, key)

def JSON_GETITEM_TEXT(json, key):
    return SQLOp("->>", json, key)

def JSON_PATH(json, keys):
    """keys is an SQL array"""
    return SQLOp("#>", json, PGArray(keys))

def JSON_PATH_TEXT(json, keys):
    """keys is an SQL array"""
    return SQLOp("#>>", json, PGArrayLiteral(keys))

def JSONB_SUPERSET(superset, subset):
    return SQLOp("@>", superset, subset)

def JSONB_SUBSET(subset, superset):
    return SQLOp("<@", subset, superset)

def JSONB_CONTAINS(jsonb, key):
    return SQLOp("?", jsonb, key)

def JSONB_CONTAINS_ANY(jsonb, keys):
    """keys is an AdditionSQL array"""
    return SQLOp("?|", jsonb, PGArray(keys))

def JSONB_CONTAINS_ALL(jsonb, keys):
    """keys is an SQL array"""
    return SQLOp("?&", jsonb, PGArray(keys))

def ARRAY_CONTAINS(arr, values):
    return SQLOp("@>", arr, PGArray(values))

def ARRAY_OVERLAPS(arr, values):
    return SQLOp("&&", arr, PGArray(values))


class JGET(object):
    """JSON field getter that JSONifies the second argument of comparisons.

    Normally it just gets a JSON key of a table field:

       >>> print(JGET("data", "key", table="Person").__sqlrepr__('postgres'))
       ((Person.data) -> ('key'))

    We can also pass a field object and omit the table:

       >>> print(JGET(Field("Person", "data"), "key").__sqlrepr__('postgres'))
       ((Person.data) -> ('key'))

    The right operand for comparison operators gets converted to JSON:

       >>> print((JGET("data", "key", table="Person") == {'foo': 'bar'}
       ...     ).__sqlrepr__('postgres'))
       (((Person.data) -> ('key')) = ('{"foo": "bar"}'::jsonb))

       >>> print((JGET("data", "key", table="Person") >= [True, False, None]
       ...     ).__sqlrepr__('postgres'))
       (((Person.data) -> ('key')) >= ('[true, false, null]'::jsonb))

    But not always (is this a good idea?):
    (adamG: no it's not a good idea, because -> returns jsonb which is never NULL
    see doctest_datetime_range)

       >>> print((JGET("data", "key", table="Person") == None
       ...     ).__sqlrepr__('postgres'))
       (((Person.data) -> ('key')) = ('null'::jsonb))

       >>> print((JGET("data", "key", table="Person") != None
       ...     ).__sqlrepr__('postgres'))
       (((Person.data) -> ('key')) <> ('null'::jsonb))
    """

    def __init__(self, field, selector, table=None):
        if table is not None:
            self.field = Field(table, field)
        else:
            self.field = field
        self.selector = selector

    def __lt__(self, other):
        return SQLOp("<", self, JSONB(json.dumps(other, sort_keys=True)))
    def __le__(self, other):
        return SQLOp("<=", self, JSONB(json.dumps(other, sort_keys=True)))
    def __gt__(self, other):
        return SQLOp(">", self, JSONB(json.dumps(other, sort_keys=True)))
    def __ge__(self, other):
        return SQLOp(">=", self, JSONB(json.dumps(other, sort_keys=True)))
    def __eq__(self, other):
        return SQLOp("=", self, JSONB(json.dumps(other, sort_keys=True)))
    def __ne__(self, other):
        return SQLOp("<>", self, JSONB(json.dumps(other, sort_keys=True)))
    def __and__(self, other):
        return SQLOp("AND", self, JSONB(json.dumps(other, sort_keys=True)))
    def __rand__(self, other):
        return SQLOp("AND", JSONB(json.dumps(other, sort_keys=True), self))
    def __or__(self, other):
        return SQLOp("OR", self, JSONB(json.dumps(other, sort_keys=True)))
    def __ror__(self, other):
        return SQLOp("OR", JSONB(json.dumps(other, sort_keys=True), self))
    def __invert__(self):
        return SQLPrefix("NOT", self)

    def __sqlrepr__(self, db):
        expr = JSON_GETITEM(self.field, self.selector)
        return sqlrepr(expr, db)


class NoTables(SQLExpression):
    """A dirty hack that fools the tablesUsedSet detection"""
    def __init__(self, expr):
        self.expr = expr

    def __sqlrepr__(self, db):
        return sqlrepr(self.expr, db)

    def tablesUsedImmediate(self, db):
        return []

    def tablesUsedSet(self, db):
        return set()

    def tablesUsed(self, db):
        return {}


class Function(SQLExpression):

    def __init__(self, fname, *args):
        self.fname = fname
        self.args = args

    def __sqlrepr__(self, db):
        args = [sqlrepr(arg, db) for arg in self.args]
        return "%s(%s)" % (self.fname, ', '.join(args))


class ARRAY_TO_STRING(Function):

    def __init__(self, array, sep):
        Function.__init__(self, 'array_to_string', array, sep)


class WithSubquery(SQLExpression):
    """Represents single subquery in WITH statement
    """
    def __init__(self, name, select, parameters=None):
        self.name = name
        self.select = select
        self.parameters = parameters

    def __sqlrepr__(self, db):
        params = ""
        if self.parameters:
            params = ', '.join(self.parameters)
            params = " (%s)" % params

        return "%s%s AS ( %s )" % (self.name, params,
                                  sqlrepr(self.select, db))


class With(SQLExpression):
    def __init__(self, subqueries, select, recursive=False):
        self.subqueries = subqueries
        self.select = select
        self.recursive = recursive

    def __sqlrepr__(self, db):
        subqs = ', '.join([sqlrepr(sq, db) for sq in self.subqueries])
        rec = " RECURSIVE" if self.recursive else ""
        return "WITH%s %s %s" % (rec, subqs, sqlrepr(self.select, db))


class _BetterUnion(Union):
    # We need to enclose queries we union into parenthesis, so queries can have
    # ORDER BY clause.
    def __sqlrepr__(self, db):
        return " UNION ".join(["( %s )" % str(sqlrepr(t, db))
                              for t in self.tables])

class UnionAll(Union):
    def __sqlrepr__(self, db):
        return " UNION ALL ".join(["( %s )" % str(sqlrepr(t, db))
                                  for t in self.tables])

# We can replace Union with _BetterUnion now, so all existing code uses it
Union = _BetterUnion

class ILIKE(LIKE):
    """Case-insensitive pair of LIKE

    ILIKE is a valid postgres keyword, but not implemented by sqlobject"""
    op = "ILIKE"



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistence Schema Fields"""
import zope.interface
import zope.schema._field
import zope.schema.interfaces
import pjpersist.serialize

@zope.interface.implementer(zope.schema.interfaces.IList)
class PJSequence(zope.schema._field.AbstractCollection):
    _type = (tuple, list, pjpersist.serialize.PersistentList)

@zope.interface.implementer(zope.schema.interfaces.IDict)
class PJMapping(zope.schema._field.Dict):
    _type = (dict, pjpersist.serialize.PersistentDict)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistence Zope Containers"""
import binascii
import persistent
import transaction
import zope.component
import warnings
import weakref
from collections.abc import MutableMapping
from future.utils import viewitems

from zope.container import contained, sample
from zope.container.interfaces import IContainer

import pjpersist.sqlbuilder as sb
from pjpersist import interfaces, serialize
from pjpersist.zope import interfaces as zinterfaces
from pjpersist.mquery import Converter

USE_CONTAINER_CACHE = True


class PJContained(contained.Contained):

    _v_name = None
    _pj_name_attr = None
    _pj_name_getter = None
    _pj_name_setter = None

    _pj_parent_attr = None
    _pj_parent_getter = None
    _pj_parent_setter = None
    _v_parent = None

    @property
    def __name__(self):
        if self._v_name is None:
            if self._pj_name_attr is not None:
                self._v_name = getattr(self, self._pj_name_attr, None)
            elif self._pj_name_getter is not None:
                self._v_name = self._pj_name_getter()
        return self._v_name

    @__name__.setter
    def __name__(self, value):
        if self._pj_name_setter is not None:
            self._pj_name_setter(value)
        self._v_name = value

    @property
    def __parent__(self):
        if self._v_parent is None:
            if self._pj_parent_attr is not None:
                self._v_parent = getattr(self, self._pj_parent_attr, None)
            elif self._pj_parent_getter is not None:
                self._v_parent = self._pj_parent_getter()
        return self._v_parent

    @__parent__.setter
    def __parent__(self, value):
        if self._pj_parent_setter is not None:
            self._pj_parent_setter(value)
        self._v_parent = value


class SimplePJContainer(sample.SampleContainer, persistent.Persistent):
    _pj_remove_documents = True

    def __getstate__(self):
        state = super(SimplePJContainer, self).__getstate__()
        state['data'] = state.pop('_SampleContainer__data')
        return state

    def __setstate__(self, state):
        # pjpersist always reads a dictionary as persistent dictionary. And
        # modifying this dictionary will cause the persistence mechanism to
        # kick in. So we create a new object that we can easily modify without
        # harm.
        state = dict(state)
        state['_SampleContainer__data'] = state.pop('data', {})
        super(SimplePJContainer, self).__setstate__(state)

    def __getitem__(self, key):
        obj = super(SimplePJContainer, self).__getitem__(key)
        obj._v_name = key
        obj._v_parent = self
        return obj

    def get(self, key, default=None):
        '''See interface `IReadContainer`'''
        obj = super(SimplePJContainer, self).get(key, default)
        if obj is not default:
            obj._v_name = key
            obj._v_parent = self
        return obj

    def items(self):
        items = super(SimplePJContainer, self).items()
        for key, obj in items:
            obj._v_name = key
            obj._v_parent = self
        return items

    def values(self):
        return [v for k, v in self.items()]

    def __setitem__(self, key, obj):
        super(SimplePJContainer, self).__setitem__(key, obj)
        self._p_changed = True

    def __delitem__(self, key):
        obj = self[key]
        super(SimplePJContainer, self).__delitem__(key)
        if self._pj_remove_documents:
            self._p_jar.remove(obj)
        self._p_changed = True


@zope.interface.implementer(IContainer, zinterfaces.IPJContainer)
class PJContainer(contained.Contained,
                  persistent.Persistent,
                  MutableMapping):
    _pj_table = None
    _pj_mapping_key = 'key'
    _pj_parent_key = 'parent'
    _pj_remove_documents = True
    _pj_id_column = 'id'
    _pj_data_column = 'data'
    _pj_column_fields = (_pj_id_column, _pj_data_column)

    def __init__(self, table=None,
                 mapping_key=None, parent_key=None):
        if table:
            self._pj_table = table
        if mapping_key is not None:
            self._pj_mapping_key = mapping_key
        if parent_key is not None:
            self._pj_parent_key = parent_key

    @property
    def _pj_jar(self):
        if not hasattr(self, '_v_mdmp'):
            # If the container is in a PJ storage hierarchy, then getting
            # the datamanager is easy, otherwise we do an adapter lookup.
            if interfaces.IPJDataManager.providedBy(self._p_jar):
                return self._p_jar

            # cache result of expensive component lookup
            self._v_mdmp = zope.component.getUtility(
                    interfaces.IPJDataManagerProvider)

        return self._v_mdmp.get(None)

    def _pj_get_parent_key_value(self):
        if getattr(self, '_p_jar', None) is None:
            raise ValueError('_p_jar not found.')
        if interfaces.IPJDataManager.providedBy(self._p_jar):
            return self
        else:
            return str('zodb-' + binascii.hexlify(self._p_oid).decode('ascii'))

    def _pj_get_id_filter(self, id):
        tbl = sb.Table(self._pj_table)
        return (getattr(tbl, self._pj_id_column) == id)

    def _pj_get_pj_mapping_key_field(self):
        """Return an sqlbuilder field or JSON field getter to be used
        to get the value of the _pj_get_pj_mapping field"""
        if self._pj_mapping_key in self._pj_column_fields:
            # if it's a native column, no need to dig in JSONB
            fld = sb.Field(self._pj_table, self._pj_mapping_key)
        else:
            datafld = sb.Field(self._pj_table, self._pj_data_column)
            fld = sb.JGET(datafld, self._pj_mapping_key)
        return fld

    def _pj_get_resolve_filter(self):
        """return a filter that selects the rows of the current container"""
        queries = []
        # Make sure that we only look through objects that have the mapping
        # key. Objects not having the mapping key cannot be part of the
        # table.
        datafld = sb.Field(self._pj_table, self._pj_data_column)
        if self._pj_mapping_key is not None:
            if self._pj_mapping_key not in self._pj_column_fields:
                # if `_pj_mapping_key` is a JSONB field make sure that it
                # exists, JGET returns jsonb which is never NULL
                # if `_pj_mapping_key` is a native column, you need to make
                # sure that it exists
                queries.append(
                    sb.JSONB_CONTAINS(datafld, self._pj_mapping_key))
        # We also make want to make sure we separate the items properly by the
        # container.
        if self._pj_parent_key is not None:
            pv = self._pj_jar._writer.get_state(self._pj_get_parent_key_value())
            queries.append(sb.JGET(datafld, self._pj_parent_key) == pv)
        return sb.AND(*queries)

    def _pj_get_list_filter(self):
        return self._pj_get_resolve_filter()

    def _combine_filters(self, *qries):
        # need to work around here an <expr> AND None situation, which
        # would become <sqlexpr> AND NULL

        notnones = [q for q in qries if q is not None]
        if not notnones:
            return sb.const.TRUE
        return sb.AND(*notnones)

    @property
    def _cache(self):
        if not USE_CONTAINER_CACHE:
            return {}
        txn = transaction.manager.get()
        self._cache_autoinvalidate()
        return txn._v_pj_container_cache.setdefault(id(self), {})

    @property
    def _cache_complete(self):
        if not USE_CONTAINER_CACHE:
            return False
        txn = transaction.manager.get()
        self._cache_autoinvalidate()
        return txn._v_pj_container_cache_complete.get(id(self), False)

    def _cache_autoinvalidate(self):
        """Make sure cache is not used after object is gc'd.

        Addresses can be reused after the object is garbage collected,
        so in order to use id(self) as the key, we need to check
        whether the object is still alive with a weakref.
        """
        txn = transaction.manager.get()
        if not hasattr(txn, '_v_pj_container_cache_sentinel'):
            txn._v_pj_container_cache_sentinel = weakref.WeakValueDictionary()
        sentinel = txn._v_pj_container_cache_sentinel

        if not hasattr(txn, '_v_pj_container_cache_complete'):
            txn._v_pj_container_cache_complete = {}
        if not hasattr(txn, '_v_pj_container_cache'):
            txn._v_pj_container_cache = {}

        if not id(self) in sentinel:
            # Blow the caches for this id
            txn._v_pj_container_cache_complete.pop(id(self), None)
            txn._v_pj_container_cache.pop(id(self), None)

        # Now the id is valid!
        sentinel[id(self)] = self

    def _cache_mark_complete(self):
        txn = transaction.manager.get()
        if not hasattr(txn, '_v_pj_container_cache_complete'):
            txn._v_pj_container_cache_complete = {}
        txn._v_pj_container_cache_complete[id(self)] = True

    def _cache_get_key(self, id, doc):
        return doc[self._pj_mapping_key]

    def _locate(self, obj, id, doc):
        """Helper method that is only used when locating items that are already
        in the container and are simply loaded from PostGreSQL."""
        if obj.__name__ is None:
            obj._v_name = doc[self._pj_mapping_key]
        if obj.__parent__ is None:
            obj._v_parent = self

    def _load_one(self, id, doc, use_cache=True):
        """Get the python object from the id/doc state"""
        if use_cache:
            cache = self._cache
            obj = cache.get(self._cache_get_key(id, doc))
            if obj is not None:
                return obj
        # Create a DBRef object and then load the full state of the object.
        dbref = serialize.DBRef(self._pj_table, id, self._pj_jar.database)
        # Stick the doc into the _latest_states:
        self._pj_jar._latest_states[dbref] = doc
        obj = self._pj_jar.load(dbref)
        self._locate(obj, id, doc)
        # Add the object into the local container cache.
        if use_cache:
            cache[obj.__name__] = obj
        return obj

    def __cmp__(self, other):
        # UserDict implements the semantics of implementing comparison of
        # items to determine equality, which is not what we want for a
        # container, so we revert back to the default object comparison.
        return cmp(id(self), id(other))

    def __getitem__(self, key):
        # First check the container cache for the object.
        obj = self._cache.get(key)
        if obj is not None:
            return obj
        if self._cache_complete:
            raise KeyError(key)
        # The cache cannot help, so the item is looked up in the database.
        fld = self._pj_get_pj_mapping_key_field()
        qry = (fld == key)
        obj = self.find_one(qry)
        if obj is None:
            raise KeyError(key)
        return obj

    def _set_mapping_and_parent(self, key, value):
        # This call by itself causes the state to change _p_changed to True.
        # but make sure we set attributes before eventually inserting...
        # saves eventually one more UPDATE query
        if self._pj_mapping_key is not None:
            setattr(value, self._pj_mapping_key, key)
        if self._pj_parent_key is not None:
            setattr(value, self._pj_parent_key, self._pj_get_parent_key_value())

    def _real_setitem(self, key, value):
        self._set_mapping_and_parent(key, value)

        # Make sure the value is in the database, since we might want
        # to use its oid.
        if value._p_oid is None:
            self._pj_jar.insert(value)

        # Also add the item to the container cache.
        # _pj_jar.insert and _cache insert must be close!
        self._cache[key] = value

    def __setitem__(self, key, value):
        # When the key is None, we need to determine it.
        orig_key = key
        if key is None:
            if self._pj_mapping_key is None:
                key = self._pj_jar.createId()
            else:
                # we have _pj_mapping_key, use that attribute
                key = getattr(value, self._pj_mapping_key, None)
                if key is None:
                    key = self._pj_jar.createId()
                    setattr(value, self._pj_mapping_key, key)
        # We want to be as close as possible to using the Zope semantics.
        self._contained_setitem(self._real_setitem, key, value, orig_key)

    def _contained_setitem(self, setitemf, key, value, check_old):
        # This is based on zope.container.contained.setitem(), but we skip
        # the check if the name is in the container in case the key was just
        # generated.
        #
        # This check generates a SELECT query and causes conflicts in case
        # there are many transactions adding to the container.
        key = contained.checkAndConvertName(key)
        marker = object()
        if check_old:
            old = self.get(key, marker)
            if old is value:
                return
            if old is not marker:
                raise KeyError(key)
        value, event = contained.containedEvent(value, self, key)
        setitemf(key, value)
        if event:
            zope.event.notify(event)
            contained.notifyContainerModified(self)

    def add(self, value, key=None):
        # We are already supporting ``None`` valued keys, which prompts the key
        # to be determined here. But people felt that a more explicit
        # interface would be better in this case.
        self[key] = value

    def __delitem__(self, key):
        value = self[key]
        # First remove the parent and name from the object.
        if self._pj_mapping_key is not None:
            try:
                delattr(value, self._pj_mapping_key)
            except AttributeError:
                # Sometimes we do not control those attributes.
                pass
        if self._pj_parent_key is not None:
            try:
                delattr(value, self._pj_parent_key)
            except AttributeError:
                # Sometimes we do not control those attributes.
                pass
        # Let's now remove the object from the database.
        if self._pj_remove_documents:
            self._pj_jar.remove(value)
        # Remove the object from the container cache.
        if USE_CONTAINER_CACHE:
            self._cache.pop(key, None)
        # Send the uncontained event.
        contained.uncontained(value, self, key)

    def __contains__(self, key):
        if self._cache_complete:
            return key in self._cache

        if key in self._cache:
            return True

        fld = self._pj_get_pj_mapping_key_field()
        qry = (fld == key)

        res = self.count(qry)
        return res > 0

    def __iter__(self):
        # If the cache contains all objects, we can just return the cache keys.
        if self._cache_complete:
            return iter(self._cache)
        fld = self._pj_get_pj_mapping_key_field()
        qry = (fld != None)
        result = self.raw_find(qry, fields=(self._pj_mapping_key,))
        return iter(doc[self._pj_mapping_key] for doc in result)

    def keys(self):
        # PY3: remove the list
        return list(self.__iter__())

    def __len__(self):
        return self.count()

    def items(self):
        return list(self.iteritems())

    def iteritems(self):
        # If the cache contains all objects, we can just return the cache items
        if self._cache_complete:
            return self._cache.items()
        result = self.raw_find()
        items = [(row[self._pj_data_column][self._pj_mapping_key],
                  self._load_one(
                      row[self._pj_id_column], row[self._pj_data_column]))
                 for row in result]
        # Signal the container that the cache is now complete.
        self._cache_mark_complete()
        # Return an iterator of the items.
        return iter(items)

    def values(self):
        return [v for _, v in self.iteritems()]

    def _get_sb_fields(self, fields):
        """Return sqlbuilder columns for a SELECT query based on
        passed field names or * if no fields are passed

        Prefers native columns where available

        See doctest_PJContainer_get_sb_fields"""
        if not fields:
            res = sb.Field(self._pj_table, '*')
        else:
            datafld = sb.Field(self._pj_table, self._pj_data_column)
            res = []
            # prefer sql columns over json fields
            for name in fields:
                if name in self._pj_column_fields:
                    # SQL names are case-insensitive, so when we return the
                    # result, we want proper capitalization.
                    res.append(sb.ColumnAS(
                        sb.Field(self._pj_table, name.lower()), name))
                else:
                    if '.' not in name:
                        accessor = sb.JSON_GETITEM(datafld, name)
                    else:
                        accessor = sb.JSON_PATH(datafld, name.split("."))
                    res.append(sb.ColumnAS(accessor, name.replace('.', '_')))
        return res

    # BBB: providing support for mongo style queries
    #      anything further is just a pain, like figuring fields
    #      and keeping return values unchanged
    #      (in the means of returning the same dict)
    def convert_mongo_query(self, spec):
        warnings.warn("Using mongo queries is deprecated",
                      DeprecationWarning, stacklevel=3)
        c = Converter(self._pj_table, self._pj_data_column)
        qry = c.convert(spec)
        return qry

    def raw_find(self, qry=None, fields=(), **kwargs):
        if isinstance(qry, dict):
            qry = self.convert_mongo_query(qry)
        qry = self._combine_filters(self._pj_get_list_filter(), qry)

        # returning the cursor instead of fetchall at the cost of not closing it
        # iterating over the cursor is better and this way we expose rowcount
        # and friends
        cur = self._pj_jar.getCursor()
        fields = self._get_sb_fields(fields)
        if qry is None:
            cur.execute(sb.Select(fields, **kwargs),
                        flush_hint=[self._pj_table])
        else:
            cur.execute(sb.Select(fields, qry, **kwargs),
                        flush_hint=[self._pj_table])
        return cur

    def find(self, qry=None, **kwargs):
        # Search for matching objects.
        result = self.raw_find(qry, **kwargs)
        for row in result:
            obj = self._load_one(
                row[self._pj_id_column], row[self._pj_data_column])
            yield obj

    def raw_find_one(self, qry=None, id=None):
        if qry is None and id is None:
            raise ValueError(
                'Missing parameter, at least qry or id must be specified.')
        if isinstance(qry, dict):
            qry = self.convert_mongo_query(qry)

        if id is not None:
            qry = self._combine_filters(
                self._pj_get_resolve_filter(), qry, self._pj_get_id_filter(id))
        else:
            qry = self._combine_filters(
                self._pj_get_list_filter(), qry)

        with self._pj_jar.getCursor() as cur:
            cur.execute(sb.Select(sb.Field(self._pj_table, '*'), qry, limit=2),
                        flush_hint=[self._pj_table])
            if cur.rowcount == 0:
                return None
            if cur.rowcount > 1:
                raise ValueError('Multiple results returned.')
            return cur.fetchone()

    def find_one(self, qry=None, id=None):
        res = self.raw_find_one(qry, id)
        if res is None:
            return None
        return self._load_one(
            res[self._pj_id_column], res[self._pj_data_column])

    def count(self, qry=None):
        if isinstance(qry, dict):
            qry = self.convert_mongo_query(qry)

        where = self._combine_filters(self._pj_get_list_filter(), qry)
        count = sb.func.COUNT(sb.Field(self._pj_table, self._pj_id_column))
        if where is None:
            select = sb.Select(count)
        else:
            select = sb.Select(count, where=where)

        with self._pj_jar.getCursor() as cur:
            cur.execute(select, flush_hint=[self._pj_table])
            return cur.fetchone()[0]

    def clear(self):
        # why items? it seems to be better to bulk-load all objects that going
        # to be deleted with one SQL query, because __delitem__ will anyway
        # load state, but then with one query for each object
        for key, value in self.items():
            del self[key]
        # Signal the container that the cache is now complete.
        # we just removed all objects, eh?
        self._cache.clear()
        self._cache_mark_complete()

    def __nonzero__(self):
        where = self._pj_get_list_filter() or True
        select = sb.Select(sb.func.COUNT(
            sb.Field(self._pj_table, self._pj_id_column)), where=where)
        with self._pj_jar.getCursor() as cur:
            cur.execute(select, flush_hint=[self._pj_table])
            return cur.fetchone()[0] > 0


class IdNamesPJContainer(PJContainer):
    """A container that uses the PostGreSQL table UID as the name/key."""
    _pj_mapping_key = None

    def __init__(self, table=None, parent_key=None):
        super(IdNamesPJContainer, self).__init__(table, parent_key)

    @property
    def _pj_remove_documents(self):
        # Objects must be removed, since removing the id of a document is not
        # allowed.
        return True

    def _cache_get_key(self, id, doc):
        return id

    def _locate(self, obj, id, doc):
        obj._v_name = id
        obj._v_parent = self

    def __getitem__(self, key):
        # First check the container cache for the object.
        obj = self._cache.get(key)
        if obj is not None:
            return obj
        if self._cache_complete:
            raise KeyError(key)
        # We do not have a cache entry, so we look up the object.
        filter = self._pj_get_resolve_filter()
        obj = self.find_one(qry=filter, id=key)
        if obj is None:
            raise KeyError(key)
        return obj

    def __contains__(self, key):
        # If all objects are loaded, we can look in the local object cache.
        if self._cache_complete:
            return key in self._cache
        # Look in PostGreSQL.
        return self.raw_find_one(id=key) is not None

    def __iter__(self):
        # If the cache contains all objects, we can just return the cache keys.
        if self._cache_complete:
            return iter(self._cache)
        # Look up all ids in PostGreSQL.
        result = self.raw_find(None, fields=(self._pj_id_column,))

        return iter(row[self._pj_id_column] for row in result)

    def iteritems(self):
        # If the cache contains all objects, we can just return the cache items
        if self._cache_complete:
            return viewitems(self._cache)
        # Load all objects from the database.
        result = self.raw_find()
        items = [(row[self._pj_id_column],
                  self._load_one(
                      row[self._pj_id_column], row[self._pj_data_column]))
                 for row in result]
        # Signal the container that the cache is now complete.
        self._cache_mark_complete()
        # Return an iterator of the items.
        return iter(items)

    def _real_setitem(self, key, value):
        # set these before inserting
        self._set_mapping_and_parent(key, value)

        # We want JSONB document ids to be our keys, so pass it to insert(), if
        # key is provided
        if value._p_oid is None:
            self._pj_jar.insert(value, key)

        # Also add the item to the container cache.
        # _pj_jar.insert and _cache insert must be close!
        self._cache[key] = value

        # no need for super, _set_mapping_and_parent does the job,
        # updating mapping and parent BEFORE inserting saves one SQL query
        # super(IdNamesPJContainer, self)._real_setitem(key, value)


class AllItemsPJContainer(PJContainer):
    _pj_parent_key = None


class SubDocumentPJContainer(PJContained, PJContainer):
    _p_pj_sub_object = True



import datetime
import pprint
import random


DTIMES = [
    datetime.datetime(2013,1,22,18,59),
    datetime.datetime(1987,2,2,8,31),
    datetime.datetime(1949,6,25,23,50),
    datetime.datetime(1918,3,3,8,35),
    datetime.datetime(1980,7,23,8,6),
    datetime.datetime(2011,11,1,16,26),
    datetime.datetime(1911,1,17,4,24),
    datetime.datetime(1975,7,22,4,44),
    datetime.datetime(1914,5,17,9,37),
    None,
    datetime.datetime(1936,12,16,10,38),
    datetime.datetime(1930,12,12,3,58),
    datetime.datetime(1995,1,23,9,59),
    datetime.datetime(1923,6,2,15,5),
    datetime.datetime(1967,3,1,13,42),
    datetime.datetime(1966,7,22,17,30),
    datetime.datetime(2005,7,13,17,18),
    datetime.datetime(1912,1,9,15,4),
    datetime.datetime(1983,6,27,3,56),
    None,
]

WORDSFILE = '/usr/share/dict/words'
WORDS = [line.strip() for line in open(WORDSFILE, 'r')]

TYPES = ('c', 'dt', 'i', 'b', 'd', 'l')
DEPTH = 0
MAXDEPTH = 5


def get_c():
    key = "'"
    while "'" in key:
        key = random.choice(WORDS)
    return key


def get_dt():
    return random.choice(DTIMES)


def get_i():
    return random.randint(0, 10000)


def get_b():
    return random.choice([True, False])


def get_d(size=5):
    global DEPTH
    DEPTH += 1
    rv = {}
    if DEPTH <= MAXDEPTH:
        for i in xrange(size):
            key = get_c()
            rv[key] = get_random_type()
    DEPTH -= 1
    return rv


def get_l():
    global DEPTH
    DEPTH += 1
    if DEPTH <= MAXDEPTH:
        ln = random.randint(0, 5)
        rv = [get_random_type() for i in xrange(ln)]
    else:
        rv = []
    DEPTH -= 1
    return rv


def get_random_type():
    value_type = random.choice(TYPES)
    getter = globals()['get_%s' % value_type]
    value = getter()
    return value


def main():
    #print '{'

    v = get_d(200)
    pprint.pprint(v)

    #for i in xrange(10):
    #    key = get_c()
    #    value_type = random.choice(TYPES)
    #    getter = globals()['get_%s' % value_type]
    #    value = getter()
    #    print "  '%s': %r," % (key, value)
    #
    #print '}'

if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Mapping Implementations"""
import json
from collections.abc import MutableMapping

from pjpersist import serialize, interfaces


class PJTableMapping(MutableMapping, object):
    __pj_table__ = None
    __pj_mapping_key__ = 'key'

    def __init__(self, jar):
        self._pj_jar = jar

    def __pj_filter__(self):
        return 'true'

    def __getitem__(self, key):
        filter = self.__pj_filter__()
        filter += ''' AND data @> '%s' ''' % json.dumps({self.__pj_mapping_key__: key})
        with self._pj_jar.getCursor() as cur:
            cur.execute(
                'SELECT id FROM ' + self.__pj_table__ + ' WHERE ' + filter)
            if not cur.rowcount:
                raise KeyError(key)
            id = cur.fetchone()['id']
            dbref = serialize.DBRef(self.__pj_table__, id, self._pj_jar.database)
        return self._pj_jar.load(dbref)

    def __setitem__(self, key, value):
        # Even though setting the attribute should register the object with
        # the data manager, the value might not be in the DB at all at this
        # point, so registering it manually ensures that new objects get added.
        self._pj_jar.register(value)
        setattr(value, interfaces.TABLE_ATTR_NAME, self.__pj_table__)
        setattr(value, self.__pj_mapping_key__, key)

    def __delitem__(self, key):
        # Deleting the object from the database is not our job. We simply
        # remove it from the dictionary.
        value = self[key]
        setattr(value, self.__pj_mapping_key__, None)

    def keys(self):
        filter = self.__pj_filter__()
        filter += """ AND NOT (data @> '{"%s": null}' OR""" % \
            self.__pj_mapping_key__
        filter += "      NOT data ?& array['%s'] )" % self.__pj_mapping_key__
        with self._pj_jar.getCursor() as cur:
            cur.execute(
                'SELECT * FROM ' + self.__pj_table__ + ' WHERE ' + filter)
            return [
                res['data'][self.__pj_mapping_key__]
                for res in cur.fetchall()]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mongo Persistence Doc Tests"""
import datetime
import doctest
import os
import unittest
from pprint import pprint

from zope.exceptions import exceptionformatter

from pjpersist import testing


class ReprMixin(object):
    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self)


def setUp(test):
    testing.setUp(test)
    # silence this, otherwise half-baked objects raise exceptions
    # on trying to __repr__ missing attributes
    test.orig_DEBUG_EXCEPTION_FORMATTER = \
        exceptionformatter.DEBUG_EXCEPTION_FORMATTER
    exceptionformatter.DEBUG_EXCEPTION_FORMATTER = 0

    def fetchone(table, clause=None):
        qry = 'SELECT * FROM ' + table
        if clause:
            qry += ' WHERE ' + clause
        with test.globs['dm'].getCursor(False) as cur:
            cur.execute(qry)
            return cur.fetchone()
    test.globs['fetchone'] = fetchone


def tearDown(test):
    testing.tearDown(test)
    exceptionformatter.DEBUG_EXCEPTION_FORMATTER = \
        test.orig_DEBUG_EXCEPTION_FORMATTER


def setUpRST(test):
    # add more stuff to globals to have less cruft around in README.rst
    setUp(test)
    test.globs['datetime'] = datetime
    test.globs['pprint'] = pprint
    test.globs['ReprMixin'] = ReprMixin


def test_suite():
    # Try to support README.rst both in-place tests and installed in a
    # virtualenv.  With a virtualenv, it drops as a data file into
    # pjpersist/README.rst relative of the virtualenv's root.
    readme_rst_path = (
        '../../../README.rst'
        if os.path.exists(os.path.join(
                os.path.dirname(__file__), '../../../README.rst'))
        else '../../../../../pjpersist/README.rst')
    suite = unittest.TestSuite((
        doctest.DocFileSuite(
            '../README.txt',
            setUp=setUp, tearDown=tearDown,
            checker=testing.checker,
            optionflags=testing.OPTIONFLAGS),
        doctest.DocFileSuite(
            readme_rst_path,
            setUp=setUpRST, tearDown=tearDown,
            checker=testing.checker,
            optionflags=testing.OPTIONFLAGS),
        ))
    suite.layer = testing.db_layer
    return suite
